/*
** Copyright (C) 1998-1999 Greg Stein. All Rights Reserved.
**
** By using this file, you agree to the terms and conditions set forth in
** the LICENSE.html file which can be found at the top level of the mod_dav
** distribution or at http://www.webdav.org/mod_dav/license-1.html.
**
** Contact information:
**   Greg Stein, PO Box 3151, Redmond, WA, 98073
**   gstein@lyra.org, http://www.webdav.org/mod_dav/
*/

/*
** DAV extension module for Apache 1.3.*
**  - Database support using DBM-style databases
**
** Writen by Greg Stein, gstein@lyra.org, http://www.lyra.org/
*/

#ifdef DAV_USE_GDBM
#include <gdbm.h>
#else
#include <fcntl.h>		/* for O_RDONLY, O_WRONLY */
#include "sdbm/sdbm.h"
#endif

#include "mod_dav.h"


#ifdef DAV_USE_GDBM

typedef GDBM_FILE dav_dbm_file;

#define DAV_DBM_CLOSE(f)	gdbm_close(f)
#define DAV_DBM_FETCH(f, k)	gdbm_fetch((f), (k))
#define DAV_DBM_STORE(f, k, v)	gdbm_store((f), (k), (v), GDBM_REPLACE)
#define DAV_DBM_DELETE(f, k)	gdbm_delete((f), (k))
#define DAV_DBM_FIRSTKEY(f)	gdbm_firstkey(f)
#define DAV_DBM_NEXTKEY(f, k)	gdbm_nextkey((f), (k))
#define DAV_DBM_CLEARERR(f)	if (0) ; else
#define DAV_DBM_FREEDATUM(f, d)	((d).dptr ? free((d).dptr) : 0)

#else

typedef DBM *dav_dbm_file;

#define DAV_DBM_CLOSE(f)	sdbm_close(f)
#define DAV_DBM_FETCH(f, k)	sdbm_fetch((f), (k))
#define DAV_DBM_STORE(f, k, v)	sdbm_store((f), (k), (v), DBM_REPLACE)
#define DAV_DBM_DELETE(f, k)	sdbm_delete((f), (k))
#define DAV_DBM_FIRSTKEY(f)	sdbm_firstkey(f)
#define DAV_DBM_NEXTKEY(f, k)	sdbm_nextkey(f)
#define DAV_DBM_CLEARERR(f)	sdbm_clearerr(f)
#define DAV_DBM_FREEDATUM(f, d)	if (0) ; else

#endif

struct dav_db {
    dav_dbm_file file;
};

#define D2G(d)	(*(datum*)&(d))


static int dav_dbm_open(pool * p,
			const char *dirpath, const char *fname,
			int ro,
			dav_db **pdb)
{
    const char *pathname;
    dav_dbm_file file;

    *pdb = NULL;

    /* Note: we assume that the caller ensured the state directory exists */
    pathname = ap_pstrcat(p,
			  dirpath,
			  "/" DAV_STATE_DIR "/",
			  fname ? fname : DAV_STATE_FILE_FOR_DIR,
			  NULL);

    /* ### readers cannot open while a writer has this open; we should
       ### perform a few retries with random pauses. */

    /* ### do we need to deal with the umask? */

    /* NOTE: stupid cast to get rid of "const" on the pathname */
#ifdef DAV_USE_GDBM
    file = gdbm_open((char *) pathname,
		     0,
		     ro ? GDBM_READER : GDBM_WRCREAT,
		     DAV_MODE_FILE,
		     NULL);
#else
    file = sdbm_open((char *) pathname,
		     ro ? O_RDONLY : (O_RDWR | O_CREAT),
		     DAV_MODE_FILE);
#endif

    /* we can't continue if we couldn't open the file and we need to write */
    if (file == NULL && !ro) {
	/* note that an internal (*DBM) error occurred */
	return DAV_DBERR_INTERNAL;
    }

    /* may be NULL if we tried to open a non-existent db as read-only */
    if (file != NULL) {
	/* we have an open database... return it */
	*pdb = ap_pcalloc(p, sizeof(**pdb));
	(*pdb)->file = file;
    }

    return DAV_DBERR_NONE;
}

static void dav_dbm_close(dav_db *db)
{
    DAV_DBM_CLOSE(db->file);
}

static int dav_dbm_fetch(dav_db *db, dav_datum key, dav_datum *pvalue)
{
    *(datum *) pvalue = DAV_DBM_FETCH(db->file, D2G(key));

    /* we don't need the error; we have *pvalue to tell */
    DAV_DBM_CLEARERR(db->file);

    return DAV_DBERR_NONE;
}

static int dav_dbm_store(dav_db *db, dav_datum key, dav_datum value)
{
    int rv;

    rv = DAV_DBM_STORE(db->file, D2G(key), D2G(value));

    /* ### fetch more specific error information? */

    /* we don't need the error; we have rv to tell */
    DAV_DBM_CLEARERR(db->file);

    return rv == -1 ? DAV_DBERR_INTERNAL : DAV_DBERR_NONE;
}

static int dav_dbm_delete(dav_db *db, dav_datum key)
{
    int rv;

    rv = DAV_DBM_DELETE(db->file, D2G(key));

    /* ### fetch more specific error information? */

    /* we don't need the error; we have rv to tell */
    DAV_DBM_CLEARERR(db->file);

    return rv == -1 ? DAV_DBERR_INTERNAL : DAV_DBERR_NONE;
}

static int dav_dbm_exists(dav_db *db, dav_datum key)
{
    int exists;

#ifdef DAV_USE_GDBM
    exists = gdbm_exists(db->file, D2G(key));
#else
    {
	datum value = sdbm_fetch(db->file, D2G(key));
	sdbm_clearerr(db->file);	/* unneeded */
	exists = value.dptr != NULL;
    }
#endif
    return exists ? DAV_DBERR_NONE : DAV_DBERR_NO_EXIST;
}

static int dav_dbm_firstkey(dav_db *db, dav_datum *pkey)
{
    *(datum *) pkey = DAV_DBM_FIRSTKEY(db->file);

    /* we don't need the error; we have *pkay to tell */
    DAV_DBM_CLEARERR(db->file);

    return DAV_DBERR_NONE;
}

static int dav_dbm_nextkey(dav_db *db, dav_datum *pkey)
{
    *(datum *) pkey = DAV_DBM_NEXTKEY(db->file, D2G(*pkey));

    /* we don't need the error; we have *pkay to tell */
    DAV_DBM_CLEARERR(db->file);

    return DAV_DBERR_NONE;
}

static int dav_dbm_geterror(dav_db *db)
{
#ifdef DAV_USE_GDBM
    return gdbm_errno;
#else
    /* ### hack. dav_propdb_error passes NULL; presume an error */
    return !db || (sdbm_error(db->file) != 0);
#endif
}

static const char *dav_dbm_strerror(dav_db *db)
{
#ifdef DAV_USE_GDBM
    return gdbm_strerror(gdbm_errno);
#else
    /* ### hack. dav_propdb_error passes NULL; presume an error */
    return (!db || sdbm_error(db->file)) ? "I/O error occurred." : "No error.";
#endif
}

static void dav_dbm_freedatum(dav_db *db, dav_datum data)
{
    DAV_DBM_FREEDATUM(db, data);
}

static dav_response dav_dbm_deleteset(pool * p,
				      const char *dirpath, const char *fname)
{
    dav_response resp =
    {0};
    char *pathname;

    /* build the propset pathname for the file */
    pathname = ap_pstrcat(p,
			  dirpath,
			  "/" DAV_STATE_DIR "/",
			  fname ? fname : DAV_STATE_FILE_FOR_DIR,
#ifndef DAV_USE_GDBM
			  DIRFEXT,
#endif
			  NULL);

    /* note: we may get ENOENT if the state dir is not present */
    if (remove(pathname) != 0 && errno != ENOENT) {
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.save_errno = errno;
	resp.desc = "Could not remove properties.";
	return resp;
    }

#ifndef DAV_USE_GDBM
    if (errno != ENOENT) {
	int extension = strlen(pathname) - 4;

	/* NOTE: we know it is 4 characters */
	memcpy(&pathname[extension], PAGFEXT, 4);

	if (remove(pathname) != 0) {
	    /* ### CRAP. only removed half. */

	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;
	    resp.desc =
		"Could not fully remove properties. "
		"The server is now in an inconsistent state.";
	    return resp;
	}
    }
#endif

    resp.status = HTTP_OK;
    return resp;
}

static dav_response dav_dbm_copymoveset(int is_move,
					pool * p,
				  const char *src_dir, const char *src_file,
				  const char *dst_dir, const char *dst_file,
					dav_buffer *pbuf)
{
    dav_response resp = { 0 };

    if (src_file == NULL)
	src_file = DAV_STATE_FILE_FOR_DIR;
    if (dst_file == NULL)
	dst_file = DAV_STATE_FILE_FOR_DIR;

#ifndef DAV_USE_GDBM
    src_file = ap_pstrcat(p, src_file, DIRFEXT, NULL);
    dst_file = ap_pstrcat(p, dst_file, DIRFEXT, NULL);
#endif

    resp = dav_copymove_state(is_move, p,
			      src_dir, src_file,
			      dst_dir, dst_file,
			      pbuf);

#ifndef DAV_USE_GDBM
    if (resp.status != HTTP_OK)
	return resp;

    {
	int extension;

	/*
	 * Revise the extension for the .PAG file. Note that we know
	 * the src/dst values are actually buffers we allocated, so the
	 * cast to lose the "const" is okay.
	 *
	 * Note that we know the extension is 4 characters.
	 */
	extension = strlen(src_file) - 4;
	memcpy((char *) &src_file[extension], PAGFEXT, 4);
	extension = strlen(dst_file) - 4;
	memcpy((char *) &dst_file[extension], PAGFEXT, 4);

	resp = dav_copymove_state(is_move, p,
				  src_dir, src_file,
				  dst_dir, dst_file,
				  pbuf);

	if (resp.status != HTTP_OK) {
	    /* ### CRAP. inconsistency. */
	    /* ### should perform some cleanup at the target if we still
	       ### have the original files */

	    /* Change the error to reflect the bad server state. */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.desc =
		"Could not fully copy/move the properties. "
		"The server is now in an inconsistent state.";
	    return resp;
	}
    }
#endif

    return resp;
}

const dav_hooks_db dav_hooks_db_dbm =
{
    dav_dbm_open,
    dav_dbm_close,
    dav_dbm_fetch,
    dav_dbm_store,
    dav_dbm_delete,
    dav_dbm_exists,
    dav_dbm_firstkey,
    dav_dbm_nextkey,
    dav_dbm_geterror,
    dav_dbm_strerror,
    dav_dbm_freedatum,
    dav_dbm_deleteset,
    dav_dbm_copymoveset,
};
