/*
** Copyright (C) 1998-1999 Greg Stein. All Rights Reserved.
**
** By using this file, you agree to the terms and conditions set forth in
** the LICENSE.html file which can be found at the top level of the mod_dav
** distribution or at http://www.webdav.org/mod_dav/license-1.html.
**
** Contact information:
**   Greg Stein, PO Box 3151, Redmond, WA, 98073
**   gstein@lyra.org, http://www.webdav.org/mod_dav/
*/

/*
** DAV extension module for Apache 1.3.*
**  - Property database handling
**
** Writen by Greg Stein, gstein@lyra.org, http://www.lyra.org/
**
** NOTES:
**
**   PROPERTY DATABASE
**
**   This version uses a SDBM or GDBM database per file and directory to
**   record the properties. These databases are kept in a subdirectory (of
**   the directory in question or the directory that holds the file in
**   question) named by the macro DAV_STATE_DIR (.DAV). The filename of the
**   database is equivalent to the target filename, and is
**   DAV_STATE_FILE_FOR_DIR (.state_for_dir) for the directory itself.
**
**   The DBM keys for the properties have the following form:
**
**     namespace ":" propname
**
**   For example: 5:author
**
**   The namespace provides an integer index into the namespace table
**   (see below). propname is simply the property name, without a namespace
**   prefix.
**
**   A special case exists for properties that had a prefix starting with
**   "xml". The XML Specification reserves these for future use. mod_dav
**   stores and retrieves them unchanged. The keys for these properties
**   have the form:
**
**     ":" propname
**
**   The propname will contain the prefix and the property name. For
**   example, a key might be ":xmlfoo:name"
**
**   The ":name" style will also be used for properties that do not
**   exist within a namespace.
**
**   The DBM values consist of two null-terminated strings, appended
**   together (the null-terms are retained and stored in the database).
**   The first string is the xml:lang value for the property. An empty
**   string signifies that a lang value was not in context for the value.
**   The second string is the property value itself.
**
**
**   NAMESPACE TABLE
**
**   The namespace table is an array that lists each of the namespaces
**   that are in use by the properties in the given propdb. Each entry
**   in the array is a simple URI.
**
**   For example: http://www.foo.bar/standards/props/
**
**   The prefix used for the property is stripped and the URI for it
**   is entered into the namespace table. Also, any namespaces used
**   within the property value will be entered into the table (and
**   stripped from the child elements).
**
**   The namespaces are stored in the DBM database under the "METADATA" key.
**
**
**   STRIPPING NAMESPACES
**
**   Within the property values, the namespace declarations (xmlns...)
**   are stripped. Each element and attribute will have its prefix removed
**   and a new prefix inserted.
**
**   This must be done so that we can return multiple properties in a
**   PROPFIND which may have (originally) used conflicting prefixes. For
**   that case, we must bind all property value elements to new namespace
**   values.
**
**   This implies that clients must NOT be sensitive to the namespace
**   prefix used for their properties. It WILL change when the properties
**   are returned (we return them as "ns<index>", e.g. "ns5"). Also, the
**   property value can contain ONLY XML elements and CDATA. PI and comment
**   elements will be stripped. CDATA whitespace will be preserved, but
**   whitespace within element tags will be altered. Attribute ordering
**   may be altered. Element and CDATA ordering will be preserved.
**
**
**   ATTRIBUTES ON PROPERTY NAME ELEMENTS
**
**   When getting/setting properties, the XML used looks like:
**
**     <prop>
**       <propname1>value</propname1>
**       <propname2>value</propname1>
**     </prop>
**
**   This implementation (mod_dav) DOES NOT save any attributes that are
**   associated with the <propname1> element. The property value is deemed
**   to be only the contents ("value" in the above example).
**
**   We do store the xml:lang value (if any) that applies to the context
**   of the <propname1> element. Whether the xml:lang attribute is on
**   <propname1> itself, or from a higher level element, we will store it
**   with the property value.
**
**
**   VERSIONING
**
**   The DBM db contains a key named "METADATA" that holds database-level
**   information, such as the namespace table. The record also contains the
**   db's version number as the very first 16-bit value. This first number
**   is actually stored as two single bytes: the first byte is a "major"
**   version number. The second byte is a "minor" number.
**
**   If the major number is not what mod_dav expects, then the db is closed
**   immediately and an error is returned. A minor number change is
**   acceptable -- it is presumed that old/new dav_props.c can deal with
**   the database format. For example, a newer dav_props might update the
**   minor value and append information to the end of the metadata record
**   (which would be ignored by previous versions).
**
**
** ISSUES:
**
**   At the moment, for the dav_get_allprops() and dav_get_props() functions,
**   we must return a set of xmlns: declarations for ALL known namespaces
**   in the file. There isn't a way to filter this because we don't know
**   which are going to be used or not. Examining property names is not
**   sufficient because the property values could use entirely different
**   namespaces.
**
**   ==> we must devise a scheme where we can "garbage collect" the namespace
**       entries from the property database.
*/

#include "mod_dav.h"

#include "http_log.h"


#define DAV_GDBM_NS_KEY		"METADATA"
#define DAV_GDBM_NS_KEY_LEN	8

#define DAV_STYLE_ISO8601	1
#define DAV_STYLE_RFC822	2

#define DAV_EMPTY_VALUE		"\0"	/* TWO null terms */

/* the namespace URI was not found; no ID is available */
#define DAV_NS_ERROR_NOT_FOUND	(DAV_NS_ERROR_BASE)

typedef struct {
    unsigned char major;
#define DAV_DBVSN_MAJOR		3
    /*
    ** V3 -- 0.9.8 ..
    **       Prior versions could have values with invalid namespace
    **       prefixes due to an incorrect mapping of input to propdb
    **       namespace indices. Version bumped to obsolete the old
    **       values.
    **
    ** V2 -- 0.9.7
    **       This introduced the xml:lang value into the property value's
    **       record in the propdb.
    **
    ** V1 -- .. 0.9.6
    **       Initial version.
    */


    unsigned char minor;
#define DAV_DBVSN_MINOR		0

    short ns_count;

} dav_propdb_metadata;

struct dav_propdb {
    int version;		/* *minor* version of this db */

    pool *p;			/* the pool we should use */

    struct stat finfo;		/* the finfo for the target file */
    dav_db *db;			/* underlying database containing props */

    dav_buffer ns_table;	/* table of namespace URIs */
    short ns_count;		/* number of entries in table */
    int ns_table_dirty;		/* ns_table was modified */

    array_header *ns_xlate;	/* translation of an elem->ns to URI */
    int *ns_map;		/* map elem->ns to propdb ns values */
};

/* list of the DAV properties (this array and the enum must stay in sync) */
static const char *dav_std_props[] =
{
    "creationdate",
    "displayname",
    "getcontentlanguage",
    "getcontentlength",
    "getcontenttype",
    "getetag",
    "getlastmodified",
    "lockdiscovery",
    "resourcetype",
    "source",
    "supportedlock",

    NULL			/* sentinel */
};
enum {
    DAV_ANNO_DAVPROP_creationdate = DAV_ANNO_DAVPROP_START,
    DAV_ANNO_DAVPROP_displayname,
    DAV_ANNO_DAVPROP_getcontentlanguage,
    DAV_ANNO_DAVPROP_getcontentlength,
    DAV_ANNO_DAVPROP_getcontenttype,
    DAV_ANNO_DAVPROP_getetag,
    DAV_ANNO_DAVPROP_getlastmodified,
    DAV_ANNO_DAVPROP_lockdiscovery,
    DAV_ANNO_DAVPROP_resourcetype,
    DAV_ANNO_DAVPROP_source,
    DAV_ANNO_DAVPROP_supportedlock,

    DAV_ANNO_DAVPROP_UNKNOWN = DAV_ANNO_DAVPROP_END
};

/*
** This structure is used to track information needed for a rollback.
** If a SET was performed and no prior value existed, then value.dptr
** will be NULL.
*/
typedef struct dav_rollback_item {
    dav_datum key;		/* key for the item being saved */
    dav_datum value;		/* value before set/replace/delete */
} dav_rollback_item;


/* ### temporary hack */
static const dav_hooks_db *hooks = &dav_hooks_db_dbm;


static int dav_find_davprop(const char *name)
{
    const char **p = dav_std_props;
    int anno = DAV_ANNO_DAVPROP_START;

    for (; *p != NULL; ++p, ++anno)
	if (strcmp(name, *p) == 0)
	    return anno;

    return DAV_ANNO_DAVPROP_UNKNOWN;
}

/* is the davprop read/write? */
static int dav_rw_davprop(dav_propdb *db, int davprop)
{
    /* resourcetype is only r/w if this isn't a collection */
    return
	davprop == DAV_ANNO_DAVPROP_displayname ||
	davprop == DAV_ANNO_DAVPROP_source ||
	davprop == DAV_ANNO_DAVPROP_UNKNOWN ||
	(davprop == DAV_ANNO_DAVPROP_resourcetype &&
	 !S_ISDIR(db->finfo.st_mode));
}

/* Note: picked up from ap_gm_timestr_822() */
/* NOTE: buf must be at least 30 chars in size */
static int dav_format_time(int style, time_t sec, char *buf)
{
    struct tm *tms;

    tms = gmtime(&sec);

    if (style == DAV_STYLE_ISO8601) {
	/* ### should we use "-00:00" instead of "Z" ?? */

	/* 20 chars plus null term */
	sprintf(buf, "%.4d-%.2d-%.2dT%.2d:%.2d:%.2dZ",
		tms->tm_year + 1900, tms->tm_mon + 1, tms->tm_mday,
		tms->tm_hour, tms->tm_min, tms->tm_sec);
	return 21;
    }

    /* RFC 822 date format; as strftime '%a, %d %b %Y %T GMT' */

    /* 29 chars plus null term */
    sprintf(buf,
	    "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
	    ap_day_snames[tms->tm_wday],
	    tms->tm_mday, ap_month_snames[tms->tm_mon],
	    tms->tm_year + 1900,
	    tms->tm_hour, tms->tm_min, tms->tm_sec);
    return 30;
}

static int dav_insert_davprop(dav_propdb *db,
			      int davprop, const char *name,
			      int getvals,
			      dav_text_header *phdr)
{
    dav_datum value = { 0 };

    /* an HTTP-date can be 29 chars plus a null term */
    /* a 64-bit size can be 20 chars plus a null term */
    char buf[30];

    switch (davprop) {
    case DAV_ANNO_DAVPROP_creationdate:
	/* closest thing to a creation date. since we don't actually
	   perform the operations that would modify ctime (after we
	   create the file), then we should be pretty safe here. */
	value.dsize = dav_format_time(DAV_STYLE_ISO8601, db->finfo.st_ctime, buf);
	value.dptr = buf;
	break;

    case DAV_ANNO_DAVPROP_getcontentlanguage:
	/* ### not yet implemented; fall through */
	break;

    case DAV_ANNO_DAVPROP_getcontentlength:
	if (S_ISREG(db->finfo.st_mode)) {
	    value.dptr = buf;
	    value.dsize = sprintf(buf, "%ld", db->finfo.st_size) + 1;
	}
	break;

    case DAV_ANNO_DAVPROP_getcontenttype:
	/* ### not yet implemented; fall through */
	break;

    case DAV_ANNO_DAVPROP_getetag:
	/* ### not yet implemented; fall through */
	break;

    case DAV_ANNO_DAVPROP_getlastmodified:
	value.dsize = dav_format_time(DAV_STYLE_RFC822, db->finfo.st_mtime, buf);
	value.dptr = buf;
	break;

    case DAV_ANNO_DAVPROP_resourcetype:
	if (S_ISDIR(db->finfo.st_mode)) {
	    /*
	     * Value for a collection. We know the "D:" prefix if provided
	     * and refers to the "DAV:" namespace URI.
	     */
	    value.dptr = "<D:collection/>";
	    value.dsize = 14;	/* 13 + null term */
	}
	else {
	    /* not handled here. this defaults to empty or whatever has been
	       stored as a dead property. */
	}
	break;

    case DAV_ANNO_DAVPROP_lockdiscovery:
    case DAV_ANNO_DAVPROP_supportedlock:
	/* ### not yet implemented; fall through */
	break;

    case DAV_ANNO_DAVPROP_displayname:
    case DAV_ANNO_DAVPROP_source:
    case DAV_ANNO_DAVPROP_UNKNOWN:
    default:
	/* fall through to interpret as a dead property */
	break;
    }

    /* if something was supplied, then insert it */
    if (value.dptr != NULL) {
	const char *s;

	if (getvals) {
	    /* use D: prefix to refer to the DAV: namespace URI */
	    s = ap_psprintf(db->p, "<D:%s>%s</D:%s>" DEBUG_CR,
			    name, value.dptr, name);
	}
	else {
	    /* use D: prefix to refer to the DAV: namespace URI */
	    s = ap_psprintf(db->p, "<D:%s/>" DEBUG_CR, name);
	}
	dav_text_append(db->p, phdr, s);

	/* we inserted a name or value (this prop is done) */
	return 1;
    }

    /* nothing was inserted */
    return 0;
}

static void dav_append_prop(dav_propdb *db,
			    const char *name, const char *value,
			    dav_text_header *phdr)
{
    const char *s;
    const char *lang = value;

    /* skip past the xml:lang value */
    value += strlen(lang) + 1;

    if (*value == '\0') {
	/* the property is an empty value */
	if (*name == ':') {
	    /* "no namespace" case */
	    s = ap_psprintf(db->p, "<%s/>" DEBUG_CR, name+1);
	}
	else {
	    s = ap_psprintf(db->p, "<ns%s/>" DEBUG_CR, name);
	}
    }
    else if (*lang != '\0') {
	if (*name == ':') {
	    /* "no namespace" case */
	    s = ap_psprintf(db->p, "<%s xml:lang=\"%s\">%s</%s>" DEBUG_CR,
			    name+1, lang, value, name+1);
	}
	else {
	    s = ap_psprintf(db->p, "<ns%s xml:lang=\"%s\">%s</ns%s>" DEBUG_CR,
			    name, lang, value, name);
	}
    }
    else if (*name == ':') {
	/* "no namespace" case */
	s = ap_psprintf(db->p, "<%s>%s</%s>" DEBUG_CR, name+1, value, name+1);
    }
    else {
	s = ap_psprintf(db->p, "<ns%s>%s</ns%s>" DEBUG_CR, name, value, name);
    }
    dav_text_append(db->p, phdr, s);
}

/*
** Prepare the ns_map variable in the db structure. This entails copying
** all URIs from the "input" namespace list (in db->ns_xlate) into the
** propdb's list of namespaces. As each URI is copied (or pre-existing
** URI looked up), the index mapping is stored into the ns_map variable.
**
** Note: we must copy all declared namespaces because we cannot easily
**   determine which input namespaces were actually used within the property
**   values that are being stored within the propdb. Theoretically, we can
**   determine this at the point where we serialize the property values
**   back into strings. This would require a bit more work, and will be
**   left to future optimizations.
**
** ### we should always initialize the propdb namespace array with "DAV:"
** ### since we know it will be entered anyhow (by virtue of it always
** ### occurring in the ns_xlate array). That will allow us to use
** ### DAV_NS_DAV_ID for propdb ns values, too.
*/
static void dav_prep_ns_map(dav_propdb *db, int add_ns)
{
    int i;
    const char **puri;
    int orig_count = db->ns_count;	/* only scan original entries */
    int *pmap;

    db->ns_map = ap_palloc(db->p, db->ns_xlate->nelts * sizeof(*db->ns_map));
    pmap = db->ns_map;

    /* ### stupid O(n * orig_count) algorithm */
    for (i = db->ns_xlate->nelts, puri = (const char **)db->ns_xlate->elts;
	 i-- > 0;
	 ++puri, ++pmap) {

	const char *uri = *puri;
	int uri_len;
	int j;
	int len;
	const char *p;

	uri_len = strlen(uri);

	p = db->ns_table.buf + sizeof(dav_propdb_metadata);

	for (j = 0; j < orig_count; ++j, p += len + 1) {
	    len = strlen(p);

	    if (uri_len != len)
		continue;
	    if (memcmp(uri, p, len) == 0) {
		*pmap = j;
		goto next_input_uri;
	    }
	}

	if (!add_ns) {
	    *pmap = DAV_NS_ERROR_NOT_FOUND;
	    continue;
	}

	dav_check_bufsize(db->p, &db->ns_table, uri_len + 1);
	memcpy(db->ns_table.buf + db->ns_table.cur_len, uri, uri_len + 1);
	db->ns_table.cur_len += uri_len + 1;

	db->ns_table_dirty = 1;

	*pmap = db->ns_count++;

   next_input_uri:
	;
    }
}

static const char *dav_get_ns_table_uri(dav_propdb *db, int ns)
{
    const char *p = db->ns_table.buf + sizeof(dav_propdb_metadata);

    while (ns--)
	p += strlen(p) + 1;

    return p;
}

/* return all known namespaces (in this propdb) */
static dav_text *dav_get_xmlns(dav_propdb *db, dav_text_header *phdr)
{
    dav_text_header hdr = { 0 };
    int i;
    const char *p = db->ns_table.buf + sizeof(dav_propdb_metadata);
    int len;

    if (phdr == NULL)
	phdr = &hdr;

    /* note: ns_count == 0 when we have no propdb file */
    for (i = 0; i < db->ns_count; ++i, p += len + 1) {
	const char *s;

	len = strlen(p);

	s = ap_psprintf(db->p, " xmlns:ns%d=\"%s\"", i, p);
	dav_text_append(db->p, phdr, s);
    }

    return phdr->first;
}

/* add a namespace decl from the input namespace table */
static void dav_add_input_xmlns(dav_propdb *db, char *marks, int ns, dav_text_header *phdr)
{
    const char *s;

    if (marks[ns])
	return;
    marks[ns] = 1;

    /* use namespaces of the form "i%d" for input namespaces */
    s = ap_psprintf(db->p, " xmlns:i%d=\"%s\"", ns, DAV_GET_NS_URI(db->ns_xlate, ns));
    dav_text_append(db->p, phdr, s);
}

/* internal function to build a key */
static dav_datum dav_gdbm_key(dav_propdb *db,
			      const dav_xml_elem *elem,
			      dav_buffer *pbuf)
{
    dav_buffer work_buf = { 0 };
    int ns;
    char nsbuf[20];
    int l_ns;
    int l_name = strlen(elem->name);
    dav_datum key = { 0 };

    if (pbuf == NULL)
	pbuf = &work_buf;

    /*
     * Convert namespace ID to a string. "no namespace" is an empty string,
     * so the keys will have the form ":name". Otherwise, the keys will
     * have the form "#:name".
     */
    if (elem->ns == DAV_NS_NONE) {
	nsbuf[0] = '\0';
	l_ns = 0;
    }
    else {
	if (db->ns_map == NULL) {
	    /*
	     * Note that we prep the map and do NOT add namespaces. If that
	     * is required, then the caller should have called prep
	     * beforehand, passing the correct values.
	     */
	    dav_prep_ns_map(db, 0);
	}

	ns = db->ns_map[elem->ns];
	if (DAV_NS_IS_ERROR(ns))
	    return key;		/* zeroed */

	l_ns = sprintf(nsbuf, "%d", ns);
    }

    /* assemble: #:name */
    dav_set_bufsize(db->p, pbuf, l_ns + 1 + l_name + 1);
    memcpy(pbuf->buf, nsbuf, l_ns);
    pbuf->buf[l_ns] = ':';
    memcpy(&pbuf->buf[l_ns + 1], elem->name, l_name + 1);

    /* build the database key */
    key.dsize = l_ns + 1 + l_name + 1;
    key.dptr = pbuf->buf;

    return key;
}


dav_propdb *dav_open_propdb(pool * p,
			    const char *dirname, const char *filename,
			    struct stat *finfo,
			    int ro,
			    array_header * ns_xlate,
			    int *perr)
{
    dav_propdb *db = ap_pcalloc(p, sizeof(*db));
    dav_datum key;
    dav_datum value = { 0 };
    int dberr;

    dberr = (*hooks->open) (p, dirname, filename, ro, &db->db);
    if (dberr != DAV_DBERR_NONE) {
	*perr = dberr;
	return NULL;
    }

    db->version = DAV_DBVSN_MINOR;
    db->p = p;
    db->finfo = *finfo;
    db->ns_xlate = ns_xlate;

    if (db->db != NULL) {
	key.dptr = DAV_GDBM_NS_KEY;
	key.dsize = DAV_GDBM_NS_KEY_LEN;
	(void) (*hooks->fetch) (db->db, key, &value);
    }
    if (value.dptr == NULL) {
	dav_propdb_metadata m =
	{DAV_DBVSN_MAJOR, DAV_DBVSN_MINOR, 0};

	if (db->db != NULL) {
	    /*
	     * If there is no METADATA key, then the database may be
	     * from versions 0.9.0 .. 0.9.4 (which would be incompatible).
	     * These can be identified by the presence of an NS_TABLE entry.
	     */
	    key.dptr = "NS_TABLE";
	    key.dsize = 8;
	    if ((*hooks->exists) (db->db, key) == DAV_DBERR_NONE) {
		(*hooks->close) (db->db);

		/* call it a major version error */
		*perr = DAV_DBERR_BAD_MAJOR;
		return NULL;
	    }
	}

	/* initialize a new metadata structure */
	dav_set_bufsize(p, &db->ns_table, sizeof(m));
	memcpy(db->ns_table.buf, &m, sizeof(m));
    }
    else {
	dav_propdb_metadata m;

	dav_set_bufsize(p, &db->ns_table, value.dsize);
	memcpy(db->ns_table.buf, value.dptr, value.dsize);

	memcpy(&m, value.dptr, sizeof(m));
	if (m.major != DAV_DBVSN_MAJOR) {
	    (*hooks->close) (db->db);

	    *perr = DAV_DBERR_BAD_MAJOR;
	    return NULL;
	}
	db->version = m.minor;
	db->ns_count = ntohs(m.ns_count);
    }

    /* ### what to do about closing the propdb on server failure? */

    return db;
}

void dav_close_propdb(dav_propdb *db)
{
    if (db->db == NULL)
	return;

    if (db->ns_table_dirty) {
	dav_propdb_metadata m;
	dav_datum key;
	dav_datum value;
	int dberr;

	key.dptr = DAV_GDBM_NS_KEY;
	key.dsize = DAV_GDBM_NS_KEY_LEN;

	value.dptr = db->ns_table.buf;
	value.dsize = db->ns_table.cur_len;

	/* fill in the metadata that we store into the prop db. */
	m.major = DAV_DBVSN_MAJOR;
	m.minor = db->version;	/* ### keep current minor version? */
	m.ns_count = htons(db->ns_count);

	memcpy(db->ns_table.buf, &m, sizeof(m));

	dberr = (*hooks->store) (db->db, key, value);
	/* ### what to do with the error? */
    }

    (*hooks->close) (db->db);
}

void dav_propdb_error(dav_propdb *db, int dberr, int *err, const char **errstr)
{
    switch (dberr) {
    case DAV_DBERR_NONE:
	*err = 0;
	*errstr = "No error.";
	break;

    case DAV_DBERR_INTERNAL:
	*err = (*hooks->geterror) (db ? db->db : NULL);
	*errstr = (*hooks->strerror) (db ? db->db : NULL);
	break;

    case DAV_DBERR_BAD_MAJOR:
	*err = DAV_DBERR_BAD_MAJOR;
	*errstr = "Incorrect major version stored in propdb.";
	break;

    case DAV_DBERR_NO_EXIST:
	*err = 0;
	*errstr = "Key does not exist.";
	break;
    }
}

dav_get_props_result dav_get_allprops(dav_propdb *db, int getvals)
{
    dav_text_header hdr = { 0 };
    dav_get_props_result result = { 0 };
    int found_resourcetype = 0;

    /* initialize the result with some start tags... */
    dav_text_append(db->p, &hdr, "<D:propstat>" DEBUG_CR "<D:prop>" DEBUG_CR);

    /* if there ARE properties, then scan them */
    if (db->db != NULL) {
	dav_datum key;

	(void) (*hooks->firstkey) (db->db, &key);
	while (key.dptr) {
	    dav_datum prevkey;
	    const char *colon;

	    /* any keys with leading capital letters should be skipped
	       (real keys start with a number or a colon) */
	    if (*key.dptr >= 'A' && *key.dptr <= 'Z')
		goto next_key;

	    /*
	     * See if this is the <DAV:resourcetype> property. We need to
	     * know whether it was found (and therefore, whether to supply
	     * a default later).
	     *
	     * Note that we've done some fast-path optimizing here:
	     *   - if first char is ':', then this is not DAV:resourcetype
	     *       (because the DAV: namespace implies a namespace ID)
	     *   - look for a single-digit namespace ID
	     *   - scan for a multiple-digit namespace ID
	     *   - if the first char of the name is 'r', then we may
	     *       have the correct property
	     */
	    if ( *key.dptr != ':' ) {
		if ( key.dptr[1] == ':' ) {
		    colon = key.dptr + 1;
		}
		else {
		    colon = strchr(key.dptr + 2, ':');
		}

		if (colon[1] == 'r'
		    && strcmp(colon + 1, "resourcetype") == 0) {

		    int ns = atoi(key.dptr);
		    const char *uri = dav_get_ns_table_uri(db, ns);

		    if (strcmp(uri, "DAV:") == 0) {
			found_resourcetype = 1;
		    }
		}
	    }

	    if (getvals) {
		dav_datum value;

		(void) (*hooks->fetch) (db->db, key, &value);
		if (value.dptr == NULL) {
		    /* ### anything better to do? */
		    /* ### probably should enter a 500 error */
		    goto next_key;
		}

		/* put the prop name and value into the result */
		dav_append_prop(db, key.dptr, value.dptr, &hdr);

		(*hooks->freedatum) (db->db, value);
	    }
	    else {
		/* simple, empty element if a value isn't needed */
		dav_append_prop(db, key.dptr, DAV_EMPTY_VALUE, &hdr);
	    }

	  next_key:
	    prevkey = key;
	    (void) (*hooks->nextkey) (db->db, &key);
	    (*hooks->freedatum) (db->db, prevkey);
	}
    }

    /* insert the standard properties */
    dav_insert_davprop(db,
		       DAV_ANNO_DAVPROP_creationdate, "creationdate",
		       getvals, &hdr);
    dav_insert_davprop(db,
		       DAV_ANNO_DAVPROP_getcontentlength, "getcontentlength",
		       getvals, &hdr);
    dav_insert_davprop(db,
		       DAV_ANNO_DAVPROP_getlastmodified, "getlastmodified",
		       getvals, &hdr);

    if (!found_resourcetype) {
	if (!dav_insert_davprop(db,
			      DAV_ANNO_DAVPROP_resourcetype, "resourcetype",
				getvals, &hdr)) {
	    /* insert the default resourcetype "value" (this also happens
	       to work for inserting only names) */
	    dav_text_append(db->p, &hdr, "<D:resourcetype/>");
	}
    }

    /* terminate the result */
    dav_text_append(db->p, &hdr,
		    "</D:prop>" DEBUG_CR
		    "<D:status>HTTP/1.1 200 OK</D:status>" DEBUG_CR
		    "</D:propstat>" DEBUG_CR);

    result.propstats = hdr.first;
    result.xmlns = dav_get_xmlns(db, NULL);
    return result;
}

dav_get_props_result dav_get_props(dav_propdb *db, dav_xml_doc *doc)
{
    dav_xml_elem *elem = dav_find_child(doc->root, "prop");
    dav_buffer buf = { 0 };
    dav_text_header hdr_good = { 0 };
    dav_text_header hdr_bad = { 0 };
    dav_text_header hdr_ns = { 0 };
    int have_good = 0;
    dav_get_props_result result = { 0 };
    char *marks;

    /* ### NOTE: we should pass in TWO buffers -- one for keys, one for
       the marks */

    /* we will ALWAYS provide a "good" result, even if it is EMPTY */
    dav_text_append(db->p, &hdr_good,
		    "<D:propstat>" DEBUG_CR
		    "<D:prop>" DEBUG_CR);

    /* generate all the namespaces that are in the propdb */
    dav_get_xmlns(db, &hdr_ns);

    /* allocate zeroed-memory for the marks. These marks indicate which
       input namespaces we've generated into the output xmlns buffer */
    marks = ap_pcalloc(db->p, db->ns_xlate->nelts);

    for (elem = elem->first_child; elem; elem = elem->next) {
	dav_datum key;
	dav_datum value = { 0 };

	if (elem->ns == DAV_NS_DAV_ID) {
	    if (elem->annotation == 0)
		elem->annotation = dav_find_davprop(elem->name);

	    /* insert the property. returns 1 if an insertion was done. */
	    if (dav_insert_davprop(db, elem->annotation, elem->name, 1, &hdr_good)) {
		have_good = 1;
		continue;
	    }
	}

	/* Note: the key may be NULL if we have no properties that are in
	   a namespace that matches the requested prop's namespace. */
	key = dav_gdbm_key(db, elem, &buf);

	/* fetch IF we have a db and a key. otherwise, value is NULL */
	if (db->db != NULL && key.dptr != NULL) {
	    (void) (*hooks->fetch) (db->db, key, &value);
	}

	/* default DAV:resourcetype to empty */
	if (value.dptr == NULL &&
	    elem->annotation == DAV_ANNO_DAVPROP_resourcetype) {
	    /* default to empty. Note that collections were forced to
	       the correct value within dav_get_davprop() */
	    dav_text_append(db->p, &hdr_good, "<D:resourcetype/>");
	    have_good = 1;
	    continue;
	}

	if (value.dptr == NULL) {
	    /* not found. add a record to the "bad" propstats */

	    /* make sure we've started our "bad" propstat */
	    if (hdr_bad.first == NULL)
		dav_text_append(db->p, &hdr_bad,
				"<D:propstat>" DEBUG_CR
				"<D:prop>" DEBUG_CR);

	    /* note: key.dptr may be NULL if the propdb doesn't have an
	       equivalent namespace stored */
	    if (key.dptr == NULL) {
		const char *s;

		if (elem->ns == DAV_NS_NONE) {
		    /*
		     * elem has a prefix already (xml...:name) or the elem
		     * simply has no namespace.
		     */
		    s = ap_psprintf(db->p, "<%s/>" DEBUG_CR, elem->name);
		}
		else {
		    /* ensure that an xmlns is generated for the
		       input namespace */
		    dav_add_input_xmlns(db, marks, elem->ns, &hdr_ns);
		    s = ap_psprintf(db->p, "<i%d:%s/>" DEBUG_CR, elem->ns, elem->name);
		}
		dav_text_append(db->p, &hdr_bad, s);
	    }
	    else {
		/* add in the bad prop using our namespace decl */
		dav_append_prop(db, key.dptr, DAV_EMPTY_VALUE, &hdr_bad);
	    }
	}
	else {
	    /* found it. put the value into the "good" propstats */

	    have_good = 1;

	    dav_append_prop(db, key.dptr, value.dptr, &hdr_good);

	    (*hooks->freedatum) (db->db, value);
	}
    }

    dav_text_append(db->p, &hdr_good,
		    "</D:prop>" DEBUG_CR
		    "<D:status>HTTP/1.1 200 OK</D:status>" DEBUG_CR
		    "</D:propstat>" DEBUG_CR);

    /* default to start with the good */
    result.propstats = hdr_good.first;

    /* we may not have any "bad" results */
    if (hdr_bad.first != NULL) {
	dav_text_append(db->p, &hdr_bad,
			"</D:prop>" DEBUG_CR
			"<D:status>HTTP/1.1 404 Not Found</D:status>" DEBUG_CR
			"</D:propstat>" DEBUG_CR);

	/* if there are no good props, then just return the bad */
	if (!have_good) {
	    result.propstats = hdr_bad.first;
	}
	else {
	    /* hook the bad propstat to the end of the good one */
	    hdr_good.last->next = hdr_bad.first;
	}
    }

    result.xmlns = hdr_ns.first;
    return result;
}

int dav_prop_exists(dav_propdb *db, dav_xml_elem *prop, dav_buffer *pbuf)
{
    dav_datum key;
    int dberr;

    if (prop->ns == DAV_NS_DAV_ID) {
	if (prop->annotation == 0)
	    prop->annotation = dav_find_davprop(prop->name);

	/* ### this isn't necessarily true, but will do for now */
	if (prop->annotation != DAV_ANNO_DAVPROP_UNKNOWN)
	    return 1;
    }

    /* no database ==> no prop */
    if (db->db == NULL)
	return 0;

    key = dav_gdbm_key(db, prop, pbuf);

    /* we know the key is not present because the namespace isn't */
    if (key.dptr == NULL)
	return 0;

    dberr = (*hooks->exists) (db->db, key);
    return dberr == DAV_DBERR_NONE;
}

dav_response dav_set_prop(dav_propdb *db, dav_xml_elem *prop, dav_buffer *pbuf)
{
    dav_response resp = { 0 };
    dav_datum key;
    dav_datum value;
    int dberr;

    if (prop->ns == DAV_NS_DAV_ID) {
	if (prop->annotation == 0)
	    prop->annotation = dav_find_davprop(prop->name);

	if (!dav_rw_davprop(db, prop->annotation)) {
	    resp.status = HTTP_CONFLICT;
	    resp.desc = "Property is read-only.";
	    return resp;
	}
    }

    /* if no database, then big time error */
    if (db->db == NULL) {
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.desc =
	    "Attempted to set a property without a valid, open "
	    "property database.";
	return resp;
    }

    /* Note: db->ns_map was set via a prior call to dav_prop_validate() */

    /* generate a text blob for the xml:lang plus the contents */
    dav_xml2text(db->p, prop, DAV_X2T_LANG_INNER, db->ns_map,
		 (const char **) &value.dptr, &value.dsize);

    key = dav_gdbm_key(db, prop, pbuf);
    dberr = (*hooks->store) (db->db, key, value);

    /* easy. stored the property okay. */
    if (dberr == DAV_DBERR_NONE) {
	resp.status = HTTP_OK;
	return resp;
    }

    /* ### we should break down some of these errors */
    resp.status = HTTP_INTERNAL_SERVER_ERROR;
    /* ### this assumes it was DAV_DBERR_INTERNAL */
    resp.desc = ap_pstrdup(db->p, (*hooks->strerror) (db->db));
    return resp;
}

dav_response dav_delete_prop(dav_propdb *db, dav_xml_elem *prop, dav_buffer *pbuf)
{
    dav_response resp = { 0 };
    dav_datum key;
    int dberr;

    if (prop->ns == DAV_NS_DAV_ID) {
	if (prop->annotation == 0)
	    prop->annotation = dav_find_davprop(prop->name);

	if (!dav_rw_davprop(db, prop->annotation)) {
	    resp.status = HTTP_CONFLICT;
	    resp.desc = "Property is read-only.";
	    return resp;
	}
    }

    /* if no database, then big time error */
    if (db->db == NULL) {
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.desc =
	    "Attempted to delete a property without a valid, open "
	    "property database.";
	return resp;
    }

    key = dav_gdbm_key(db, prop, pbuf);

    /* ### strictly speaking, this case will never occur since we
       ### do a "preflight" test using dav_prop_exists() */
    /* we may already know the key isn't present (because a namespace
       isn't present) */
    if (key.dptr == NULL) {
	resp.status = HTTP_NOT_FOUND;
	resp.desc = "The property to delete was not found.";
	return resp;
    }

    dberr = (*hooks->delete) (db->db, key);

    /* easy. deleted the property okay. */
    if (dberr == DAV_DBERR_NONE) {
	resp.status = HTTP_OK;
	return resp;
    }

    /* ### we should break down some of these errors */
    resp.status = HTTP_INTERNAL_SERVER_ERROR;
    /* ### this assumes it was DAV_DBERR_INTERNAL */
    resp.desc = ap_pstrdup(db->p, (*hooks->strerror) (db->db));
    return resp;
}

dav_response dav_delete_propset(pool * p, const char *pathname)
{
    const char *dirpath = ap_make_dirstr_parent(p, pathname);
    const char *fname = pathname + strlen(dirpath);

    /*
     * NOTE: "pathname" will never refer to a directory. We don't have
     *       have to worry about managing to pass NULL for fname.
     */

    return (*hooks->deleteset) (p, dirpath, fname);
}

dav_response dav_copymove_propset(int is_move,
				  pool * p,
				  const char *src,
				  const char *dst,
				  dav_buffer *pbuf)
{
    const char *srcdir = ap_make_dirstr_parent(p, src);
    const char *srcfile = src + strlen(srcdir);
    const char *dstdir = ap_make_dirstr_parent(p, dst);
    const char *dstfile = dst + strlen(dstdir);

    /*
     * NOTE: "src" and "dst" will never refer to a directory. We don't
     *       have to worry about managing to pass NULL for fname.
     */

    return (*hooks->copymoveset) (is_move, p,
				  srcdir, srcfile,
				  dstdir, dstfile,
				  pbuf);
}

void dav_prop_validate(dav_prop_ctx *ctx)
{
    /*
    ** ### this routine is a no-op for now, except that we need to see
    ** ### if any SET operations are being performed. If so, then we
    ** ### need to copy all the input namespaces into the propdb namespace
    ** ### list. Of course, this isn't quite right... if we have an
    ** ### error, and a rollback occurs, then the namespace mods should
    ** ### not happen at all. Basically, the namespace management is
    ** ### simply a bitch. This is a first pass solution.
    */
    if (ctx->operation == DAV_PROP_OP_SET) {
	dav_prep_ns_map(ctx->db, 1);
    }
}

void dav_prop_exec(dav_prop_ctx *ctx, dav_buffer *pbuf)
{
}

void dav_prop_commit(dav_prop_ctx *ctx)
{
}

void dav_prop_rollback(dav_prop_ctx *ctx)
{
}
