/*
** Copyright (C) 1998-1999 Greg Stein. All Rights Reserved.
**
** By using this file, you agree to the terms and conditions set forth in
** the LICENSE.html file which can be found at the top level of the mod_dav
** distribution or at http://www.webdav.org/mod_dav/license-1.html.
**
** Contact information:
**   Greg Stein, PO Box 3151, Redmond, WA, 98073
**   gstein@lyra.org, http://www.webdav.org/mod_dav/
*/

/*
** DAV extension module for Apache 1.3.*
**  - various utilities
**
** Written by Greg Stein, gstein@lyra.org, http://www.lyra.org/
*/

#include "mod_dav.h"

#include "http_request.h"
#include "http_config.h"
#include "http_vhost.h"
#include "http_log.h"


void dav_text_append(pool * p, dav_text_header *hdr, const char *text)
{
    dav_text *t = ap_palloc(p, sizeof(*t));

    t->text = text;
    t->next = NULL;

    if (hdr->first == NULL) {
	/* no text elements yet */
	hdr->first = hdr->last = t;
    }
    else {
	/* append to the last text element */
	hdr->last->next = t;
	hdr->last = t;
    }
}

void dav_check_bufsize(pool * p, dav_buffer *pbuf, int extra_needed)
{
    /* grow the buffer if necessary */
    if (pbuf->cur_len + extra_needed > pbuf->alloc_len) {
	char *newbuf;

	pbuf->alloc_len += extra_needed + DAV_BUFFER_PAD;
	newbuf = ap_palloc(p, pbuf->alloc_len);
	memcpy(newbuf, pbuf->buf, pbuf->cur_len);
	pbuf->buf = newbuf;
    }
}

void dav_set_bufsize(pool * p, dav_buffer *pbuf, int size)
{
    /* NOTE: this does not retain prior contents */

    /* NOTE: this function is used to init the first pointer, too, since
       the PAD will be larger than alloc_len (0) for zeroed structures */

    /* grow if we don't have enough for the requested size plus padding */
    if (size + DAV_BUFFER_PAD > pbuf->alloc_len) {
	/* set the new length; min of MINSIZE */
	pbuf->alloc_len = size + DAV_BUFFER_PAD;
	if (pbuf->alloc_len < DAV_BUFFER_MINSIZE)
	    pbuf->alloc_len = DAV_BUFFER_MINSIZE;

	pbuf->buf = ap_palloc(p, pbuf->alloc_len);
    }
    pbuf->cur_len = size;
}

#if APACHE_RELEASE == 10304100
/* ### this code can be used for 1.3.4 installations. use this function
 * ### instead of ap_sub_req_method_uri()
 */
/*
 * ### don't look at this code.
 * ### it is a Crime Against All That is Right and Good
 */
#include "http_core.h"		/* for SATISFY_* */
static request_rec *gross_hack(const char *new_file, const request_rec * r)
{
    request_rec *rnew = ap_sub_req_lookup_uri(new_file, r);
    int res;

    /* ### these aren't exported properly from the headers */
    extern int ap_check_access(request_rec *);	/* check access on non-auth basis */
    extern int ap_check_user_id(request_rec *);		/* obtain valid username from client auth */
    extern int ap_check_auth(request_rec *);	/* check (validated) user is authorized here */

    if (rnew->status != HTTP_OK)
	return rnew;

    /* re-run portions with a modified method */
    rnew->method = r->method;
    rnew->method_number = r->method_number;

    if ((ap_satisfies(rnew) == SATISFY_ALL
	 || ap_satisfies(rnew) == SATISFY_NOSPEC)
	? ((res = ap_check_access(rnew))
	   || (ap_some_auth_required(rnew)
	       && ((res = ap_check_user_id(rnew))
		   || (res = ap_check_auth(rnew)))))
	: ((res = ap_check_access(rnew))
	   && (!ap_some_auth_required(rnew)
	       || ((res = ap_check_user_id(rnew))
		   || (res = ap_check_auth(rnew)))))
	) {
	rnew->status = res;
    }

    return rnew;
}
#endif /* APACHE_RELEASE == 10304100 */

/*
** dav_lookup_uri()
**
** Extension for ap_sub_req_lookup_uri() which can't handle absolute
** URIs properly.
**
** If NULL is returned, then an error occurred with parsing the URI or
** the URI does not match the current server.
*/
dav_lookup_result dav_lookup_uri(const char *uri, request_rec * r)
{
    dav_lookup_result result =
    {0};
    const char *scheme;
    unsigned port = ntohs(r->connection->local_addr.sin_port);
    uri_components comp;
    char *new_file;

    /* first thing to do is parse the URI into various components */
    if (ap_parse_uri_components(r->pool, uri, &comp) != HTTP_OK) {
	result.resp.status = HTTP_BAD_REQUEST;
	result.resp.desc = "Invalid syntax in Destination URI.";
	return result;
    }

    /* the URI must be an absoluteURI (WEBDAV S9.3) */
    if (comp.scheme == NULL) {
	result.resp.status = HTTP_BAD_REQUEST;
	result.resp.desc = "Destination URI must be an absolute URI.";
	return result;
    }

    /* ### not sure this works if the current request came in via https: */
    scheme = r->parsed_uri.scheme;
    if (scheme == NULL)
	scheme = ap_http_method(r);

    /* insert a port if the URI did not contain one */
    if (comp.port == 0)
	comp.port = ap_default_port_for_scheme(comp.scheme);

    /* now, verify that the URI uses the same scheme as the current request.
       the port, must match our port.
       the URI must not have a query (args) or a fragment
     */
    if (strcasecmp(comp.scheme, scheme) != 0 ||
	comp.port != port) {
	result.resp.status = HTTP_BAD_GATEWAY;
	result.resp.desc = ap_psprintf(r->pool,
				       "Destination URI refers to different "
				       "scheme or port (%s://hostname:%d)\n"
				       "(want: %s://hostname:%d)",
				       comp.scheme ? comp.scheme : scheme,
				       comp.port ? comp.port : port,
				       scheme, port);
	return result;
    }
    if (comp.query != NULL || comp.fragment != NULL) {
	result.resp.status = HTTP_BAD_REQUEST;
	result.resp.desc =
	    "Destination URI contains invalid components "
	    "(a query or a fragment).";
	return result;
    }

    /* we have verified the scheme, port, and general structure */

    /* now, if a hostname was provided, then verify that it represents the
       same server as the current connection. note that we just use our
       port, since we've verified the URI matches ours */
    if (comp.hostname != NULL &&
	!ap_matches_request_vhost(r, comp.hostname, port)) {
	result.resp.status = HTTP_BAD_GATEWAY;
	result.resp.desc = "Destination URI refers to a different server.";
	return result;
    }

    /* we have verified that the requested URI denotes the same server as
       the current request. Therefore, we can use ap_sub_req_lookup_uri() */

    /* reconstruct a URI as just the path */
    new_file = ap_unparse_uri_components(r->pool, &comp, UNP_OMITSITEPART);

    /*
     * Lookup the URI and return the sub-request. Note that we use the
     * same HTTP method on the destination. This allows the destination
     * to apply appropriate restrictions (e.g. readonly).
     */
#if APACHE_RELEASE == 10304100
    result.rnew = gross_hack(new_file, r);
#else
    result.rnew = ap_sub_req_method_uri(r->method, new_file, r);
#endif

    return result;
}

/* ensure that our state subdirectory is present */
void dav_ensure_state_dir(pool * p, const char *dirname)
{
    const char *pathname = ap_pstrcat(p, dirname, "/" DAV_STATE_DIR, NULL);

    /* ### do we need to deal with the umask? */

    /* just try to make it, ignoring any resulting errors */
    mkdir(pathname, DAV_MODE_DIR);
}

dav_response dav_copymove_file(int is_move,
			       pool * p,
			       const char *src,
			       const char *dst,
			       dav_buffer *pbuf)
{
    dav_response resp = { 0 };
    dav_buffer work_buf = { 0 };
    int fdi;
    int fdo;

    if (pbuf == NULL)
	pbuf = &work_buf;

    dav_set_bufsize(p, pbuf, DAV_COPY_BLOCKSIZE);

    if ((fdi = open(src, O_RDONLY)) == -1) {
	/* ### use something besides 500? */
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.save_errno = errno;
	resp.desc = "Could not open file for reading";
	return resp;
    }

    /* ### do we need to deal with the umask? */
    if ((fdo = open(dst, O_WRONLY | O_CREAT | O_TRUNC, DAV_MODE_FILE)) == -1) {
	close(fdi);

	/* ### use something besides 500? */
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.save_errno = errno;
	resp.desc = "Could not open file for writing";
	return resp;
    }

    while (1) {
	ssize_t len = read(fdi, pbuf->buf, DAV_COPY_BLOCKSIZE);

	if (len == -1) {
	    close(fdi);
	    close(fdo);

	    if (remove(dst) != 0) {
		/* ### ACK! Inconsistent state... */

		/* ### use something besides 500? */
		resp.status = HTTP_INTERNAL_SERVER_ERROR;
		resp.save_errno = errno;
		resp.desc =
		    "Could not delete output after read failure. "
		    "Server is now in an inconsistent state.";
		return resp;
	    }

	    /* ### use something besides 500? */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;
	    resp.desc = "Could not read input file";
	    return resp;
	}
	if (len == 0)
	    break;

	if (write(fdo, pbuf->buf, len) != len) {
	    close(fdi);
	    close(fdo);

	    if (remove(dst) != 0) {
		/* ### ACK! Inconsistent state... */

		/* ### use something besides 500? */
		resp.status = HTTP_INTERNAL_SERVER_ERROR;
		resp.save_errno = errno;
		resp.desc =
		    "Could not delete output after write failure. "
		    "Server is now in an inconsistent state.";
		return resp;
	    }

	    /* ### use something besides 500? */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;
	    resp.desc = "Could not write output file";
	    return resp;
	}
    }

    close(fdi);
    close(fdo);

    if (is_move && remove(src) != 0) {
	/* save away the errno that got us here */
	resp.save_errno = errno;

	if (remove(dst) != 0) {
	    /* ### ACK. this creates an inconsistency. do more!? */

	    /* ### use something besides 500? */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;	/* use the new errno */
	    resp.desc =
		"Could not remove source or destination file. "
		"Server is now in an inconsistent state.";
	    return resp;
	}

	/* ### use something besides 500? */
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	/* note: we're using the saved errno */
	resp.desc =
	    "Could not remove source file after move. "
	    "Destination was removed to ensure consistency.";
	return resp;
    }

    resp.status = HTTP_OK;
    return resp;
}

/* ### need more buffers to replace the pool argument */
dav_response dav_copymove_state(int is_move,
				pool * p,
				const char *src_dir, const char *src_file,
				const char *dst_dir, const char *dst_file,
				dav_buffer *pbuf)
{
    dav_response resp = { 0 };
    struct stat src_finfo;	/* finfo for source file */
    struct stat dst_state_finfo;	/* finfo for STATE directory */
    const char *src;
    const char *dst;


    /* build the propset pathname for the source file */
    src = ap_pstrcat(p, src_dir, "/" DAV_STATE_DIR "/", src_file, NULL);

    /* the source file doesn't exist */
    if (stat(src, &src_finfo) != 0) {
	resp.status = HTTP_OK;
	return resp;
    }

    /* build the pathname for the destination state dir */
    dst = ap_pstrcat(p, dst_dir, "/" DAV_STATE_DIR, NULL);

    /* ### do we need to deal with the umask? */

    /* ensure that it exists */
    if (mkdir(dst, DAV_MODE_DIR) != 0) {
	if (errno != EEXIST) {
	    /* ### use something besides 500? */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;
	    resp.desc = "Could not create internal state directory";
	    return resp;
	}
    }

    /* get info about the state directory */
    if (stat(dst, &dst_state_finfo) != 0) {
	/* Ack! Where'd it go? */
	/* ### use something besides 500? */
	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.save_errno = errno;
	resp.desc = "State directory disappeared";
	return resp;
    }

    /* The mkdir() may have failed because a *file* exists there already */
    if (!S_ISDIR(dst_state_finfo.st_mode)) {
	/* ### try to recover by deleting this file? (and mkdir again) */
	/* ### use something besides 500? */

	resp.status = HTTP_INTERNAL_SERVER_ERROR;
	resp.save_errno = errno;
	resp.desc = "State directory is actually a file";
	return resp;
    }

    /* append the target file to the state directory pathname */
    dst = ap_pstrcat(p, dst, "/", dst_file, NULL);

    /* copy/move the file now */
    if (is_move && src_finfo.st_dev == dst_state_finfo.st_dev) {
	/* simple rename is possible since it is on the same device */
	if (rename(src, dst) != 0) {
	    /* ### use something besides 500? */
	    resp.status = HTTP_INTERNAL_SERVER_ERROR;
	    resp.save_errno = errno;
	    resp.desc = "Could not move state file.";
	    return resp;
	}
    }
    else {
	/* gotta copy (and delete) */
	return dav_copymove_file(is_move, p, src, dst, pbuf);
    }

    resp.status = HTTP_OK;
    return resp;
}


/* ---------------------------------------------------------------
**
** XML UTILITY FUNCTIONS
*/

/* validate that the root element uses a given DAV: tagname (TRUE==valid) */
int dav_validate_root(const dav_xml_doc *doc, const char *tagname)
{
    return doc->root &&
	doc->root->ns == DAV_NS_DAV_ID &&
	strcmp(doc->root->name, tagname) == 0;
}

/* find and return the (unique) child with a given DAV: tagname */
dav_xml_elem *dav_find_child(const dav_xml_elem *elem, const char *tagname)
{
    dav_xml_elem *child = elem->first_child;

    for (; child; child = child->next)
	if (child->ns == DAV_NS_DAV_ID && !strcmp(child->name, tagname))
	    return child;
    return NULL;
}


/* how many characters for the given integer? */
#define DAV_NS_LEN(ns)	((ns) < 10 ? 1 : (ns) < 100 ? 2 : (ns) < 1000 ? 3 : \
			 (ns) < 10000 ? 4 : (ns) < 100000 ? 5 : \
			 (ns) < 1000000 ? 6 : (ns) < 10000000 ? 7 : \
			 (ns) < 100000000 ? 8 : (ns) < 1000000000 ? 9 : 10)

static int dav_text_size(const dav_text *t)
{
    int size = 0;

    for (; t; t = t->next)
	size += strlen(t->text);
    return size;
}

static int dav_elem_size(const dav_xml_elem *elem, int style, int *ns_map)
{
    int size;

    if (style == DAV_X2T_FULL) {
	const dav_xml_attr *attr;

	if (elem->ns == DAV_NS_NONE) {
	    /* compute size of: <%s> */
	    size = 1 + strlen(elem->name) + 1;
	}
	else {
	    int ns = ns_map ? ns_map[elem->ns] : elem->ns;

	    /* compute size of: <ns%d:%s> */
	    size = 3 + DAV_NS_LEN(ns) + 1 + strlen(elem->name) + 1;
	}

	if (DAV_ELEM_IS_EMPTY(elem)) {
	    /* insert a closing "/" */
	    size += 1;
	}
	else {
	    /*
	     * two of above plus "/":
	     *     <ns%d:%s> ... </ns%d:%s>
	     * OR  <%s> ... </%s>
	     */
	    size = 2 * size + 1;
	}

	for (attr = elem->attr; attr; attr = attr->next) {
	    if (attr->ns == DAV_NS_NONE) {
		/* compute size of: ' %s="%s"' */
		size += 1 + strlen(attr->name) + 2 + strlen(attr->value) + 1;
	    }
	    else {
		/* compute size of: ' ns%d:%s="%s"' */
		size += 3 + DAV_NS_LEN(attr->ns) + 1 + strlen(attr->name) + 2 + strlen(attr->value) + 1;
	    }
	}

	/*
	 * If the element has an xml:lang value that is *different* from
	 * its parent, then add the thing in: ' xml:lang="%s"'. Note that
	 * we take advantage of the pointer equality established by the
	 * parsing for "inheriting" the xml:lang values from parents.
	 */
	if (elem->lang != NULL &&
	    (elem->parent == NULL || elem->lang != elem->parent->lang)) {
	    size += 11 + strlen(elem->lang) + 1;
	}
    }
    else if (style == DAV_X2T_LANG_INNER) {
	/*
	 * This style prepends the xml:lang value plus a null terminator.
	 * If a lang value is not present, then we insert a null term.
	 */
	size = elem->lang ? strlen(elem->lang) + 1 : 1;
    }
    else
	size = 0;

    size += dav_text_size(elem->first_cdata.first);

    for (elem = elem->first_child; elem; elem = elem->next) {
	/* the size of the child element plus the CDATA that follows it */
	size += (dav_elem_size(elem, DAV_X2T_FULL, ns_map) +
		 dav_text_size(elem->following_cdata.first));
    }

    return size;
}

static char *dav_write_text(char *s, const dav_text *t)
{
    for (; t; t = t->next) {
	int len = strlen(t->text);
	memcpy(s, t->text, len);
	s += len;
    }
    return s;
}

static char *dav_write_elem(char *s, const dav_xml_elem *elem, int style,
			    int *ns_map)
{
    const dav_xml_elem *child;
    int len;
    int ns;

    if (style == DAV_X2T_FULL) {
	int empty = DAV_ELEM_IS_EMPTY(elem);
	const dav_xml_attr *attr;

	if (elem->ns == DAV_NS_NONE) {
	    len = sprintf(s, "<%s", elem->name);
	}
	else {
	    ns = ns_map ? ns_map[elem->ns] : elem->ns;
	    len = sprintf(s, "<ns%d:%s", ns, elem->name);
	}
	s += len;

	for (attr = elem->attr; attr; attr = attr->next) {
	    if (attr->ns == DAV_NS_NONE)
		len = sprintf(s, " %s=\"%s\"", attr->name, attr->value);
	    else
		len = sprintf(s, " ns%d:%s=\"%s\"", attr->ns, attr->name, attr->value);
	    s += len;
	}

	/* add the xml:lang value if necessary */
	if (elem->lang != NULL &&
	    (elem->parent == NULL || elem->lang != elem->parent->lang)) {
	    len = sprintf(s, " xml:lang=\"%s\"", elem->lang);
	    s += len;
	}

	/* no more to do. close it up and go. */
	if (empty) {
	    *s++ = '/';
	    *s++ = '>';
	    return s;
	}

	/* just close it */
	*s++ = '>';
    }
    else if (style == DAV_X2T_LANG_INNER) {
	/* prepend the xml:lang value */
	if (elem->lang != NULL) {
	    len = strlen(elem->lang);
	    memcpy(s, elem->lang, len);
	    s += len;
	}
	*s++ = '\0';
    }

    s = dav_write_text(s, elem->first_cdata.first);

    for (child = elem->first_child; child; child = child->next) {
	s = dav_write_elem(s, child, DAV_X2T_FULL, ns_map);
	s = dav_write_text(s, child->following_cdata.first);
    }

    if (style == DAV_X2T_FULL) {
	if (elem->ns == DAV_NS_NONE) {
	    len = sprintf(s, "</%s>", elem->name);
	}
	else {
	    ns = ns_map ? ns_map[elem->ns] : elem->ns;
	    len = sprintf(s, "</ns%d:%s>", ns, elem->name);
	}
	s += len;
    }

    return s;
}

/* convert an element to a text string */
void dav_xml2text(pool * p,
		  const dav_xml_elem *elem,
		  int style,
		  int *ns_map,
		  const char **pbuf,
		  int *psize)
{
    int size = dav_elem_size(elem, style, ns_map) + 1;	/* null term */
    char *s = ap_palloc(p, size);

    dav_write_elem(s, elem, style, ns_map);
    s[size - 1] = '\0';

    *pbuf = s;
    if (psize)
	*psize = size;
}

const char *dav_empty_elem(pool * p, const dav_xml_elem *elem)
{
    if (elem->ns == DAV_NS_NONE) {
	/*
	 * The prefix (xml...) is already within the prop name, or
	 * the element simply has no prefix.
	 */
	return ap_psprintf(p, "<%s/>" DEBUG_CR, elem->name);
    }

    return ap_psprintf(p, "<ns%d:%s/>" DEBUG_CR, elem->ns, elem->name);
}
