/*
** Copyright (C) 1998-1999 Greg Stein. All Rights Reserved.
**
** By using this file, you agree to the terms and conditions set forth in
** the LICENSE.html file which can be found at the top level of the mod_dav
** distribution or at http://www.webdav.org/mod_dav/license-1.html.
**
** Contact information:
**   Greg Stein, PO Box 3151, Redmond, WA, 98073
**   gstein@lyra.org, http://www.webdav.org/mod_dav/
*/

/*
** DAV extension module for Apache 1.3.*
**
** Writen by Greg Stein, gstein@lyra.org, http://www.lyra.org/
**
** APACHE ISSUES:
**   - within a DAV hierarchy, if an unknown method is used and we default
**     to Apache's implementation, it sends back an OPTIONS with the wrong
**     set of methods -- there is NO HOOK for us.
**     therefore: we need to manually handle the HTTP_METHOD_NOT_ALLOWED
**       and HTTP_NOT_IMPLEMENTED responses (not ap_send_error_response).
**   - process_mkcol_body() had to dup code from ap_setup_client_block().
**   - it would be nice to get status lines from Apache for arbitrary
**     status codes
**   - it would be nice to be able to extend Apache's set of response
**     codes so that it doesn't return 500 when an unknown code is placed
**     into r->status.
**   - http_vhost functions should apply "const" to their params
**
** DESIGN NOTES:
**   - For PROPFIND, we batch up the entire response in memory before
**     sending it. We may want to reorganize around sending the information
**     as we suck it in from the propdb. Alternatively, we should at least
**     generate a total Content-Length if we're going to buffer in memory
**     so that we can keep the connection open.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include "mod_dav.h"


#define DAV_INFINITY	INT_MAX	/* for the Depth: header */

/* ### may need to tweak this a bit: we don't want to allow the extra "/"
 * ### if we're looking at a file (i.e. disallow ../file.html/ ) */
#define DAV_REQ_HAS_EXTRA_PATH_INFO(r) \
    (*(r)->path_info != '\0' \
     && (*(r)->path_info != '/' \
         || (r)->path_info[1] != '\0'))

/* method handler hooks for other modules */
typedef struct dav_hooks {
    dav_api_request_handler request_handler;
    dav_api_request_handler mkcol_test;
    dav_api_request_handler mkcol_create;

#if 0
    dav_api_method_handler put_handler;
    dav_api_method_handler delete_handler;
    dav_api_method_handler options_handler;
    dav_api_method_handler propfind_handler;
    dav_api_method_handler proppatch_handler;
    dav_api_method_handler copy_handler;
    dav_api_method_handler move_handler;
    dav_api_method_handler lock_handler;
    dav_api_method_handler unlock_handler;
#endif
} dav_hooks;

/* per-dir configuration */
typedef struct {
    int enabled;
    const char *dir;
    dav_hooks *hooks;		/* ### ptr or direct embedding? */
} dav_dir_conf;

/* per-server configuration */
typedef struct {
    table *content_type;	/* content-type to array of grammars */
    table *query_grammar;	/* query grammar name to module name */
    table *coded_uri;		/* query grammar name to URI */
} dav_server_conf;


/* directory tree walking context */
typedef struct dav_walker_ctx {
    int walk_type;
#define DAV_WALKTYPE_AUTH	1	/* walk authorized files */
#define DAV_WALKTYPE_FULL	2	/* walk ALL files */

    int split_path;		/* split dir/fname? */
    int postfix;		/* call func for dirs after files */

    int (*func)(struct dav_walker_ctx *ctx);

    pool *pool;

    dav_buffer pathbuf;
    const char *fname;
    struct stat finfo;

    dav_buffer uri;		/* current URI */
    dav_buffer path2;		/* optional secondary path */

    dav_response *response;

    /* for PROPFIND operations */
    dav_xml_doc *doc;
    int propfind_type;
#define DAV_PROPFIND_IS_ALLPROP		1
#define DAV_PROPFIND_IS_PROPNAME	2
#define DAV_PROPFIND_IS_PROP		3

    dav_get_props_result badprops;	/* cached propstat of unfound props */

    /* for COPY and MOVE operations */
    int is_move;
    int overwrite;
    int move_props;
    dav_buffer work_buf;

    /* ### temporary hack for dav_copymove_walker to log errors */
    request_rec *r;

} dav_walker_ctx;

/* forward-declare for use in configuration lookup */
module MODULE_VAR_EXPORT dav_module;

static void dav_init_handler(server_rec *s, pool *p)
{
    ap_add_version_component("DAV/" DAV_VERSION);
}

static void *dav_create_server_config(pool *p, server_rec *s)
{
    dav_server_conf *newconf;

    newconf = (dav_server_conf *) ap_pcalloc(p, sizeof(*newconf));

    newconf->content_type = ap_make_table(p, 1);
    newconf->query_grammar = ap_make_table(p, 1);
    newconf->coded_uri = ap_make_table(p, 1);

    return newconf;
}

static void *dav_merge_server_config(pool *p, void *base, void *overrides)
{
    dav_server_conf *parent = base;
    dav_server_conf *child = overrides;
    dav_server_conf *newconf;

    newconf = (dav_server_conf *) ap_pcalloc(p, sizeof(*newconf));
    newconf->content_type = ap_copy_table(p, parent->content_type);
    newconf->query_grammar = ap_copy_table(p, parent->query_grammar);
    newconf->coded_uri = ap_copy_table(p, parent->coded_uri);

    /* ### ap_overlap_tables relies on qsort() ... is that a stable sort? */

    ap_overlap_tables(newconf->content_type, child->content_type,
		      AP_OVERLAP_TABLES_SET);
    ap_overlap_tables(newconf->query_grammar, child->query_grammar,
		      AP_OVERLAP_TABLES_SET);
    ap_overlap_tables(newconf->coded_uri, child->coded_uri,
		      AP_OVERLAP_TABLES_SET);

    return newconf;
}

static void *dav_create_dir_config(pool *p, char *dir)
{
    /* NOTE: dir==NULL creates the default per-dir config */

    dav_dir_conf *conf;

    conf = (dav_dir_conf *) ap_pcalloc(p, sizeof(*conf));
    conf->dir = ap_pstrdup(p, dir);
    conf->hooks = (dav_hooks *) ap_pcalloc(p, sizeof(dav_hooks));

    return conf;
}

static void *dav_merge_dir_config(pool *p, void *base, void *overrides)
{
    dav_dir_conf *parent = base;
    dav_dir_conf *child = overrides;
    dav_dir_conf *newconf = (dav_dir_conf *) ap_pcalloc(p, sizeof(*newconf));

    newconf->enabled = parent->enabled || child->enabled;
    if (parent->hooks) {
	newconf->hooks = parent->hooks;
    }
    else {
	newconf->hooks = child->hooks;
    }
    return newconf;
}

/*
 * Command handler for the DAV directive, which is FLAG.
 */
static const char *dav_cmd_dav(cmd_parms *cmd, void *config, int arg)
{
    dav_dir_conf *conf = (dav_dir_conf *) config;

    conf->enabled = arg;
    return NULL;
}

/*
 * Command handler for the AddQueryGrammar directive, which is TAKE3.
 */
static const char *dav_cmd_addquerygrammar(cmd_parms *cmd, void *dummy,
					   char *arg1, char *arg2,
					   char *arg3)
{
    server_rec *s = cmd->server;
    dav_server_conf *conf;

    conf = (dav_server_conf *) ap_get_module_config(s->module_config,
						    &dav_module);

    /* ### we will probably want to have a bit more data per grammar; this
     * ### would change to a mapping of grammar names to structs */
    ap_table_set(conf->query_grammar, arg1, arg2);
    ap_table_set(conf->coded_uri, arg1, arg3);

    return NULL;
}

/*
 * Command handler for the AllowQueryGrammar directive, which is ITERATE2.
 */
static const char *dav_cmd_allowquerygrammar(cmd_parms *cmd, void *dummy,
					     char *arg1, char *arg2)
{
    server_rec *s = cmd->server;
    dav_server_conf *conf;
    array_header *grammars;

    conf = (dav_server_conf *) ap_get_module_config(s->module_config,
						    &dav_module);
    grammars = (array_header *) ap_table_get(conf->content_type, arg1);
    if (grammars == NULL) {
	grammars = ap_make_array(cmd->pool, 5, sizeof(const char *));
	ap_table_addn(conf->content_type, arg1, (const char *) grammars);
    }

    /* ### we should look this value up and store a ptr to our internal
     * ### structure (whatever that will be) */
    *(const char **) ap_push_array(grammars) = arg2;

    return NULL;
}

static int dav_error_response(request_rec *r, int status, const char *sline,
			      const char *body)
{
    r->status = status;
    if (sline != NULL) {	/* we provide when Apache doesn't have one */
	r->status_line = (char *) sline;	/* ### crap!! */
    }
    r->content_type = "text/html";

    ap_send_http_header(r);

    /* the status_line gets filled in by ap_send_http_header() */
    if (sline == NULL) {
	sline = r->status_line;
    }

    /* ### hard or soft? */
    ap_soft_timeout("send error body", r);

    ap_rvputs(r,
	      DAV_RESPONSE_BODY_1,
	      sline,
	      DAV_RESPONSE_BODY_2,
	      &sline[4],
	      DAV_RESPONSE_BODY_3,
	      NULL);

    ap_rputs(body, r);

    ap_rputs(ap_psignature("\n<P><HR>\n", r), r);
    ap_rputs(DAV_RESPONSE_BODY_4, r);

    ap_kill_timeout(r);

    /* the response has been sent. */
    /*
     * ### Use of DONE obviates logging..!
     */
    return DONE;
}

/* handy function for return values of methods that (may) create things */
static int dav_created(request_rec *r, request_rec *rnew, const char *what)
{
    const char *body;

    if (rnew == NULL) {
	rnew = r;
    }

    /* did the target file already exist? */
    if (rnew->finfo.st_mode != 0) {
	/* Apache will supply a default message */
	return HTTP_NO_CONTENT;
    }

    /* Per HTTP/1.1, S10.2.2: add a Location header to contain the
     * URI that was created. */

    /* ### rnew->uri does not contain an absoluteURI. S14.30 states that
     * ### the Location header requires an absoluteURI. where to get it? */
    /* ### disable until we get the right value */
#if 0
    ap_table_setn(r->headers_out, "Location", rnew->uri);
#endif

    /* ### insert an ETag header? see HTTP/1.1 S10.2.2 */

    /* Apache doesn't allow us to set a variable body for HTTP_CREATED, so
     * we must manufacture the entire response. */
    body = ap_psprintf(r->pool, "%s %s has been created.",
		       what,
		       ap_escape_html(rnew->pool, rnew->uri));
    return dav_error_response(r, HTTP_CREATED, NULL, body);
}

static int dav_get_depth(request_rec *r, int def_depth)
{
    const char *depth = ap_table_get(r->headers_in, "Depth");

    if (depth == NULL) {
	return def_depth;
    }
    if (strcasecmp(depth, "infinity") == 0) {
	return DAV_INFINITY;
    }
    else if (strcmp(depth, "0") == 0) {
	return 0;
    }
    else if (strcmp(depth, "1") == 0) {
	return 1;
    }

    /* The caller will return an HTTP_BAD_REQUEST. This will augment the
     * default message that Apache provides. */
    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		  "An invalid Depth header was specified.");
    return -1;
}

static int dav_get_overwrite(request_rec *r)
{
    const char *overwrite = ap_table_get(r->headers_in, "Overwrite");

    if (overwrite == NULL) {
	return 1;		/* default is "T" */
    }

    if (strcmp(overwrite, "F") == 0) {
	return 0;
    }
    if (strcmp(overwrite, "T") == 0) {
	return 1;
    }

    /* The caller will return an HTTP_BAD_REQUEST. This will augment the
     * default message that Apache provides. */
    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		  "An invalid Overwrite header was specified.");
    return -1;
}

/* handle the PUT method */
static int dav_method_put(request_rec *r)
{
    const char *body;
    int result;
    FILE *f;

    if ((result = ap_setup_client_block(r, REQUEST_CHUNKED_DECHUNK)) != OK) {
	return result;
    }

    /* ### appropriate lock checks go in here somewhere... */

    /* ### we probably don't want to allow this "/" on the end since PUT
     * ### can't really go to a directory, and we don't want to see a URL
     * ### that looks like /some/path/file.html/
     */
    if (DAV_REQ_HAS_EXTRA_PATH_INFO(r)) {
	/* a parent directory did not exist */

	/* Create some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Missing one or more intermediate collections. "
			   "Cannot create resource %s.",
			   ap_escape_html(r->pool, r->uri));
	return dav_error_response(r, HTTP_CONFLICT, NULL, body);
    }
    if (S_ISDIR(r->finfo.st_mode)) {
	/* Create some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Cannot PUT to a collection.");
	return dav_error_response(r, HTTP_CONFLICT, NULL, body);
    }

    f = ap_pfopen(r->pool, r->filename, "w");
    if (f == NULL) {
	/* Apache will supply a default error for this. */
	/* ### not quite right... we should mention PUT */
	return HTTP_FORBIDDEN;
    }

    if (ap_should_client_block(r)) {
	char *buffer = ap_palloc(r->pool, DAV_READ_BLOCKSIZE);
	int len;

	while ((len = ap_get_client_block(r, buffer,
					  DAV_READ_BLOCKSIZE)) > 0) {
	    fwrite(buffer, len, 1, f);
	}

	/* ### if len == -1, then an error occurred. what to do? */
    }

    ap_pfclose(r->pool, f);

    /* NOTE: WebDAV spec, S8.7.1 states properties should be unaffected */

    /* return an appropriate response (HTTP_CREATED or HTTP_NO_CONTENT) */
    return dav_created(r, NULL, "Resource");
}

static void dav_add_response(dav_walker_ctx *ctx, const char *href, int status,
			     dav_get_props_result *propstats)
{
    dav_response *resp;

    /* just drop some data into an dav_response */
    resp = ap_pcalloc(ctx->pool, sizeof(*resp));
    resp->href = ap_pstrdup(ctx->pool, href);
    resp->status = status;
    if (propstats) {
	resp->propstats = *propstats;
    }
    resp->next = ctx->response;

    ctx->response = resp;
}

static int dav_walker(dav_walker_ctx *ctx, int depth)
{
    int isdir = S_ISDIR(ctx->finfo.st_mode);
    DIR *dirp;
    struct dirent *ep;
    int result = OK;
    char *slash;

    /* ensure the context is prepared properly, then call the func */
    if (isdir) {
	ctx->fname = NULL;
	result = (*ctx->func)(ctx);
    }
    else {
	slash = strrchr(ctx->pathbuf.buf, '/');

	if (ctx->split_path) {
	    /* Note: this could result in the dirname being "" */
	    *slash = '\0';
	}
	ctx->fname = slash + 1;
	result = (*ctx->func)(ctx);
	*slash = '/';
    }
    if (result != OK) {
	return result;
    }

    if (depth == 0 || !isdir) {
	return OK;
    }

    /* put a trailing slash onto the directory, in preparation for appending
     * files to it as we discovery them within the directory */
    dav_check_bufsize(ctx->pool, &ctx->pathbuf, DAV_BUFFER_PAD);
    ctx->pathbuf.buf[ctx->pathbuf.cur_len++] = '/';
    ctx->pathbuf.buf[ctx->pathbuf.cur_len] = '\0';	/* in pad area */

    /* Note: the URI should ALREADY have a trailing "/" */

    /* if a secondary path is present, then do that, too */
    if (ctx->path2.buf != NULL) {
	dav_check_bufsize(ctx->pool, &ctx->path2, DAV_BUFFER_PAD);
	ctx->path2.buf[ctx->path2.cur_len++] = '/';
	ctx->path2.buf[ctx->path2.cur_len] = '\0';	/* in pad area */
    }

    /* open and scan the directory */
    if ((dirp = opendir(ctx->pathbuf.buf)) == NULL) {
#if 0
	/* ### debug hack */
	dav_add_response(ctx, "no directory", HTTP_NOT_FOUND, NULL);
	return OK;
#endif

	/* ### need a better error */
	return HTTP_NOT_FOUND;
    }
    while ((ep = readdir(dirp)) != NULL) {
	int len = strlen(ep->d_name);

	/* avoid recursing into our current, parent, or state directories */
	if (ep->d_name[0] == '.'
	    && (len == 1 || (ep->d_name[1] == '.' && len == 2))) {
	    continue;
	}

	if (ctx->walk_type == DAV_WALKTYPE_AUTH) {
	    /* ### need to authorize each file */
	    /* ### example: .htaccess is normally configured to fail auth */

	    if (!strcmp(ep->d_name, DAV_STATE_DIR)) {
		continue;
	    }
	}

	/* grow the buffer if necessary */
	dav_check_bufsize(ctx->pool, &ctx->pathbuf, len + 1);

	/* append this file onto the path buffer (copy null term) */
	memcpy(ctx->pathbuf.buf + ctx->pathbuf.cur_len, ep->d_name, len + 1);

	if (stat(ctx->pathbuf.buf, &ctx->finfo) != 0) {
#if 0
	    /* ### debug hack */
	    dav_add_response(ctx, "file disappeared", HTTP_NOT_FOUND, NULL);
	    dav_add_response(ctx, ctx->pathbuf.buf, HTTP_NOT_FOUND, NULL);
	    return OK;
#endif

	    /* woah! where'd it go? */
	    result = HTTP_NOT_FOUND;
	    break;
	}

	/* copy the file to the URI, too. NOTE: we will check for the
	 * length, a null term, and a (possible) trailing slash for a dir */
	dav_check_bufsize(ctx->pool, &ctx->uri, len + 2);
	memcpy(ctx->uri.buf + ctx->uri.cur_len, ep->d_name, len + 1);

	/* if there is a secondary path, then do that, too */
	if (ctx->path2.buf != NULL) {
	    dav_check_bufsize(ctx->pool, &ctx->path2, len + 1);
	    memcpy(ctx->path2.buf + ctx->path2.cur_len, ep->d_name, len + 1);
	}

	/* ### for now, only process regular files (e.g. skip symlinks) */
	if (S_ISREG(ctx->finfo.st_mode)) {
	    slash = &ctx->pathbuf.buf[ctx->pathbuf.cur_len - 1];

	    /* break up the pathname into directory + filename */
	    if (ctx->split_path) {
		*slash = '\0';
	    }
	    ctx->fname = &ctx->pathbuf.buf[ctx->pathbuf.cur_len];

	    /* call the function for the specified dir + file */
	    if ((result = (*ctx->func)(ctx)) != OK) {
		break;
	    }

	    /* restore the slash on the directory */
	    *slash = '/';
	}
	else if (S_ISDIR(ctx->finfo.st_mode)) {
	    int save_path_len = ctx->pathbuf.cur_len;
	    int save_uri_len = ctx->uri.cur_len;
	    int save_path2_len = ctx->path2.cur_len;

	    /* adjust length to incorporate the subdir name */
	    ctx->pathbuf.cur_len += len;
	    ctx->path2.cur_len += len;

	    /* adjust URI length to incorporate subdir and a slash */
	    ctx->uri.cur_len += len + 1;
	    ctx->uri.buf[ctx->uri.cur_len - 1] = '/';
	    ctx->uri.buf[ctx->uri.cur_len] = '\0';

	    /* recurse on the subdir */
	    if ((result = dav_walker(ctx, depth - 1)) != OK) {
		break;
	    }

	    ctx->pathbuf.cur_len = save_path_len;
	    ctx->uri.cur_len = save_uri_len;
	    ctx->path2.cur_len = save_path2_len;
	}
    }

    /* ### check the return value of this? */
    closedir(dirp);

    if (result == OK && ctx->postfix) {
	/* replace the dirs' trailing slashes with null terms */
	ctx->pathbuf.buf[--ctx->pathbuf.cur_len] = '\0';
	ctx->uri.buf[--ctx->uri.cur_len] = '\0';
	if (ctx->path2.buf != NULL) {
	    ctx->path2.buf[--ctx->path2.cur_len] = '\0';
	}

	ctx->fname = NULL;
	ctx->finfo.st_mode = 0;	/* mark finfo as invalid */
	result = (*ctx->func)(ctx);
    }

    return result;
}

static int dav_delete_walker(dav_walker_ctx *ctx)
{
    /* ### what is the best way to log an error? */

    if (ctx->fname != NULL) {
	/* call for a specific file */

	if (remove(ctx->pathbuf.buf) != 0) {
	}
    }
    else if (ctx->finfo.st_mode == 0) {
	/* postfix call on directory */

	if (rmdir(ctx->pathbuf.buf) != 0) {
	}
    }

    return OK;
}

static int dav_delete_tree(request_rec *r)
{
    dav_walker_ctx ctx = { 0 };

    ctx.walk_type = DAV_WALKTYPE_FULL;
    ctx.postfix = 1;
    ctx.func = dav_delete_walker;
    ctx.pool = r->pool;
    ctx.finfo = r->finfo;

    /* copy over the filename (with null term) */
    dav_set_bufsize(r->pool, &ctx.pathbuf, strlen(r->filename));
    memcpy(ctx.pathbuf.buf, r->filename, ctx.pathbuf.cur_len + 1);

    /* copy over the URI (with null term) */
    dav_set_bufsize(r->pool, &ctx.uri, strlen(r->uri));
    memcpy(ctx.uri.buf, r->uri, ctx.uri.cur_len + 1);

    return dav_walker(&ctx, DAV_INFINITY);
}

/* handle the DELETE method */
static int dav_method_delete(request_rec *r)
{
    dav_response resp;
    int result;

    if ((result = ap_discard_request_body(r)) != OK) {
	return result;
    }

    if (r->finfo.st_mode == 0) {
	/* Apache will supply a default error for this. */
	return HTTP_NOT_FOUND;
    }

    if (dav_get_depth(r, DAV_INFINITY) != DAV_INFINITY) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "Depth must be \"infinity\" for DELETE.");
	return HTTP_BAD_REQUEST;
    }

    /* ### locking crap goes in here somewhere... */

    if (S_ISDIR(r->finfo.st_mode)) {
	if ((result = dav_delete_tree(r)) != OK) {
	    return result;
	}
    }
    else if (remove(r->filename) != 0) {
	/* Apache will supply a default error for this. */
	/* ### not quite right. we should mention DELETE */
	return HTTP_FORBIDDEN;
    }

    /* toss the file's properties */
    resp = dav_delete_propset(r->pool, r->filename);
    if (resp.status != HTTP_OK) {
	if (resp.desc != NULL) {
	    errno = resp.save_errno;
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s", resp.desc);
	}
	return resp.status;
    }

    /* ### HTTP_NO_CONTENT if no body, HTTP_OK if there is a body (some day) */

    /* Apache will supply a default error for this. */
    return HTTP_NO_CONTENT;
}

/* handle the OPTIONS method */
static int dav_method_options(request_rec *r)
{
    int result;

    /* per HTTP/1.1 S9.2, we can discard this body */
    if ((result = ap_discard_request_body(r)) != OK) {
	return result;
    }

    /* no body */
    ap_set_content_length(r, 0);

    /* ### we probably need subsets of the methods here! */
    ap_table_setn(r->headers_out, "Allow",
		  "GET, HEAD, POST, PUT, DELETE, OPTIONS, TRACE, "
		  "PROPFIND, PROPPATCH, MKCOL, COPY, MOVE"
#if 0
    /* ### when we support this */
		  ", LOCK, UNLOCK"
#endif
	);

    /* ### change to "1,2" when we add locking */
    ap_table_setn(r->headers_out, "DAV", "1");

    /* ### this will send a Content-Type. the default OPTIONS does not. */
    ap_send_http_header(r);

    /* ### the default (ap_send_http_options) returns OK, but I believe
     * ### that is because it is the default handler and nothing else
     * ### will run after the thing. */

    /* we've sent everything necessary to the client. */
    return DONE;
}

static void dav_send_multistatus(request_rec *r, dav_response *first,
				 array_header *namespaces)
{
    ap_rputs(DAV_XML_HEADER DEBUG_CR
	     "<D:multistatus xmlns:D=\"DAV:\"", r);

    if (namespaces != NULL) {
	int i;

	for (i = namespaces->nelts; i--; ) {
	    ap_rprintf(r, " xmlns:ns%d=\"%s\"", i,
		       DAV_GET_NS_URI(namespaces, i));
	}
    }

    /* ap_rputc('>', r); */
    ap_rputs(">" DEBUG_CR, r);

    while (first) {
	dav_text *t;

	if (first->propstats.xmlns == NULL) {
	    ap_rputs("<D:response>", r);
	}
	else {
	    ap_rputs("<D:response", r);
	    for (t = first->propstats.xmlns; t; t = t->next) {
		ap_rputs(t->text, r);
	    }
	    ap_rputc('>', r);
	}

	ap_rputs(DEBUG_CR "<D:href>", r);
	ap_rputs(first->href, r);	/* ### need quoting! */
	ap_rputs("</D:href>" DEBUG_CR, r);

	if (first->propstats.propstats == NULL) {
	    /* ### it would be nice to get a status line from Apache */
	    ap_rprintf(r,
		       "<D:status>HTTP/1.1 %d status text goes here</D:status>"
		       DEBUG_CR, first->status);
	}
	else {
	    /* assume this includes <propstat> and is quoted properly */
	    for (t = first->propstats.propstats; t; t = t->next) {
		ap_rputs(t->text, r);
	    }
	}

	if (first->desc != NULL) {
	    ap_rputs("<D:responsedescription>", r);
	    ap_rputs(first->desc, r);	/* ### need quoting! */
	    ap_rputs("</D:responsedescription>" DEBUG_CR, r);
	}

	ap_rputs("</D:response>" DEBUG_CR, r);

	first = first->next;
    }

    ap_rputs("</D:multistatus>" DEBUG_CR, r);
}

/* ### debug hack */
static void recurse_elem(dav_xml_doc *doc, dav_xml_elem *elem, request_rec *r)
{
    dav_xml_attr *attr = elem->attr;
    dav_xml_elem *child = elem->first_child;

    if (elem->ns == DAV_NS_NONE) {
	ap_rprintf(r, "<%s", elem->name);
    }
    else {
	ap_rprintf(r, "<ns%d:%s", elem->ns, elem->name);
    }

    for ( ; attr; attr = attr->next) {
	if (attr->ns == DAV_NS_NONE) {
	    ap_rprintf(r, " %s=\"%s\"", attr->name, attr->value);
	}
	else {
	    ap_rprintf(r, " ns%d:%s=\"%s\"", attr->ns, attr->name,
		       attr->value);
	}
    }

    if (elem->parent == NULL) {
	int i;

	for (i = doc->namespaces->nelts; i--; ) {
	    ap_rprintf(r, " xmlns:ns%d=\"%s\"", i,
		       DAV_GET_NS_URI(doc->namespaces, i));
	}
    }

    if (DAV_ELEM_IS_EMPTY(elem)) {
	ap_rputs("/>", r);
    }
    else {
	dav_text *t;

	ap_rputc('>', r);

	for (t = elem->first_cdata.first; t; t = t->next) {
	    ap_rputs(t->text, r);
	}

	for ( ; child; child = child->next) {
	    recurse_elem(doc, child, r);
	    for (t = child->following_cdata.first; t; t = t->next) {
		ap_rputs(t->text, r);
	    }
	}

	if (elem->ns == DAV_NS_NONE) {
	    ap_rprintf(r, "</%s>", elem->name);
	}
	else {
	    ap_rprintf(r, "</ns%d:%s>", elem->ns, elem->name);
	}
    }
}

static void dav_cache_badprops(dav_walker_ctx *ctx)
{
    const dav_xml_elem *elem;
    dav_text_header hdr = { 0 };

    /* just return if we built the thing already */
    if (ctx->badprops.propstats != NULL) {
	return;
    }

    dav_text_append(ctx->pool, &hdr,
		    "<D:propstat>" DEBUG_CR
		    "<D:prop>" DEBUG_CR);

    elem = dav_find_child(ctx->doc->root, "prop");
    for (elem = elem->first_child; elem; elem = elem->next) {
	dav_text_append(ctx->pool, &hdr, dav_empty_elem(ctx->pool, elem));
    }

    dav_text_append(ctx->pool, &hdr,
		    "</D:prop>" DEBUG_CR
		    "<D:status>HTTP/1.1 404 Not Found</D:status>" DEBUG_CR
		    "</D:propstat>" DEBUG_CR);

    ctx->badprops.propstats = hdr.first;
}

static int dav_propfind_walker(dav_walker_ctx *ctx)
{
    const char *dirname = ctx->pathbuf.buf;
    dav_propdb *db;
    dav_get_props_result propstats = { 0 };
    int dberr;

    /* rare case for root of the filesystem(!) */
    if (*dirname == '\0') {
	dirname = "/";
    }

    /* Note: ctx->doc can only be NULL for DAV_PROPFIND_IS_ALLPROP. Since
     * dav_get_allprops() does not need to do namespace translation,
     * we're okay. */
    db = dav_open_propdb(ctx->pool, dirname, ctx->fname, &ctx->finfo, 1,
			 ctx->doc ? ctx->doc->namespaces : NULL, &dberr);
    if (db == NULL) {
	/* ### do something with dberr! */

	if (ctx->propfind_type == DAV_PROPFIND_IS_PROP) {
	    /* some props were expected on this collection/resource */
	    dav_cache_badprops(ctx);
	    dav_add_response(ctx, ctx->uri.buf, HTTP_NOT_FOUND,
			     &ctx->badprops);
	}
	else {
	    /* no props on this collection/resource */
	    dav_add_response(ctx, ctx->uri.buf, HTTP_OK, NULL);
	}
	return OK;
    }
    /* ### what to do about closing the propdb on server failure? */

    if (ctx->propfind_type == DAV_PROPFIND_IS_PROP) {
	propstats = dav_get_props(db, ctx->doc);
    }
    else {
	propstats = dav_get_allprops(db,
				     ctx->propfind_type
				     == DAV_PROPFIND_IS_ALLPROP);
    }
    dav_close_propdb(db);

    dav_add_response(ctx, ctx->uri.buf, HTTP_OK, &propstats);

    return OK;
}

/* handle the PROPFIND method */
static int dav_method_propfind(request_rec *r)
{
    int depth;
    int result;
    dav_xml_doc *doc;
    const dav_xml_elem *child;
    dav_walker_ctx ctx = { 0 };


    if (r->finfo.st_mode == 0) {
	/* Apache will supply a default error for this. */
	return HTTP_NOT_FOUND;
    }
    if ((depth = dav_get_depth(r, DAV_INFINITY)) < 0) {
	/* dav_get_depth() supplies additional information for the
	 * default message. */
	return HTTP_BAD_REQUEST;
    }

    if ((result = dav_parse_input(r, &doc)) != OK) {
	return result;
    }
    /* note: doc == NULL if no request body */

    if (doc && !dav_validate_root(doc, "propfind")) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "The \"propfind\" element was not found.");
	return HTTP_BAD_REQUEST;
    }

    /* ### validate that only one of these three elements is present */

    if (doc == NULL
	|| (child = dav_find_child(doc->root, "allprop")) != NULL) {
	/* note: no request body implies allprop */
	ctx.propfind_type = DAV_PROPFIND_IS_ALLPROP;
    }
    else if ((child = dav_find_child(doc->root, "propname")) != NULL) {
	ctx.propfind_type = DAV_PROPFIND_IS_PROPNAME;
    }
    else if ((child = dav_find_child(doc->root, "prop")) != NULL) {
	ctx.propfind_type = DAV_PROPFIND_IS_PROP;
    }
    else {
	/* "propfind" element must have one of the above three children */

	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "The \"propfind\" element does not contain one of "
		      "the required child elements (the specific command).");
	return HTTP_BAD_REQUEST;
    }

    ctx.walk_type = DAV_WALKTYPE_AUTH;
    ctx.split_path = 1;
    ctx.func = dav_propfind_walker;
    ctx.pool = r->pool;
    ctx.doc = doc;
    ctx.finfo = r->finfo;

    /* copy the filename into the buffer (with null term) */
    /* Note: null term occurs within pad area of buffer */
    dav_set_bufsize(r->pool, &ctx.pathbuf, strlen(r->filename));
    memcpy(ctx.pathbuf.buf, r->filename, ctx.pathbuf.cur_len + 1);

    /* copy over the URI (with null term) */
    dav_set_bufsize(r->pool, &ctx.uri, strlen(r->uri));
    memcpy(ctx.uri.buf, r->uri, ctx.uri.cur_len + 1);

    /* if we have a directory, then ensure the URI has a trailing "/" */
    if (S_ISDIR(r->finfo.st_mode)
	&& ctx.uri.buf[ctx.uri.cur_len - 1] != '/') {
	/* this will fall into the pad area we allocated */
	ctx.uri.buf[ctx.uri.cur_len++] = '/';
	ctx.uri.buf[ctx.uri.cur_len] = '\0';
    }

    if ((result = dav_walker(&ctx, depth)) != OK) {
	return result;
    }

    /* all PROPFIND results are "207 Multi-Status" and of type text/xml */
    r->status = HTTP_MULTI_STATUS;
    r->content_type = DAV_XML_CONTENT_TYPE;

    ap_send_http_header(r);

    ap_soft_timeout("send PROPFIND response", r);

    /* if a badprops was generated, then we need to spit out the doc's
     * namespaces for use by the badprops. Note that <response> elements
     * will override these ns0, ns1, etc, but NOT within the <response>
     * scope for the badprops. */
    /* NOTE: badprops.propstats != NULL implies doc != NULL */
    if (ctx.badprops.propstats == NULL) {
	dav_send_multistatus(r, ctx.response, NULL);
    }
    else {
	dav_send_multistatus(r, ctx.response, doc->namespaces);
    }

#if 0
    if (doc && doc->root) {
	recurse_elem(doc, doc->root, r);
    }
#endif

    ap_kill_timeout(r);

    /* the response has been sent. */
    return DONE;
}

static dav_response dav_failed_proppatch(pool *p, dav_xml_doc *doc)
{
    dav_xml_elem *child;
    dav_text_header hdr_404 = { 0 };
    dav_text_header hdr_424 = { 0 };
    dav_response resp = { 0 };

    /* we know we'll always have a 404 section ... it's why we're here */
    dav_text_append(p, &hdr_404,
		    "<D:propstat>" DEBUG_CR
		    "<D:prop>" DEBUG_CR);

    for (child = doc->root->first_child; child; child = child->next) {
	const dav_xml_elem *prop;
	dav_xml_elem *one_prop;

	/* Ignore children that are not set/remove */
	if (child->annotation == DAV_ANNO_IGNORE) {
	    continue;
	}

	/* we know this is present */
	prop = dav_find_child(child, "prop");

	for (one_prop = prop->first_child; one_prop;
	     one_prop = one_prop->next) {
	    dav_text_header *phdr;

	    if (one_prop->annotation == DAV_ANNO_PROPPATCH_404) {
		phdr = &hdr_404;
	    }
	    else {
		phdr = &hdr_424;
	    }

	    if (phdr->first == NULL) {
		dav_text_append(p, phdr,
				"<D:propstat>" DEBUG_CR
				"<D:prop>" DEBUG_CR);
	    }

	    dav_text_append(p, phdr, dav_empty_elem(p, one_prop));
	}
    }

    dav_text_append(p, &hdr_404,
		    "</D:prop>" DEBUG_CR
		    "<D:status>HTTP/1.1 404 Not Found</D:status>" DEBUG_CR
		    "</D:propstat>" DEBUG_CR);

    /* if a 424 section was generated, then append it */
    if (hdr_424.first != NULL) {
	dav_text_append(p, &hdr_424,
			"</D:prop>" DEBUG_CR
			"<D:status>HTTP/1.1 424 Failed Dependency</D:status>"
			DEBUG_CR
			"</D:propstat>" DEBUG_CR);
	hdr_404.last->next = hdr_424.first;

    }

    resp.propstats.propstats = hdr_404.first;
    return resp;
}

static dav_response dav_exec_proppatch(pool *p, dav_xml_doc *doc,
				       dav_propdb *db, dav_buffer *pbuf)
{
    dav_xml_elem *child;
    dav_text_header hdr = { 0 };
    dav_response resp = { 0 };

    dav_text_append(p, &hdr,
		    "<D:propstat>" DEBUG_CR
		    "<D:prop>" DEBUG_CR);

    for (child = doc->root->first_child; child; child = child->next) {
	int is_remove;
	const dav_xml_elem *prop;
	dav_xml_elem *one_prop;

	/* Ignore children that are not set/remove */
	if (child->annotation == DAV_ANNO_IGNORE) {
	    continue;
	}

	/* we left a marker to help us */
	is_remove = child->annotation == DAV_ANNO_DAVTAG_REMOVE;

	/* we know this is present */
	prop = dav_find_child(child, "prop");

	/* iterate over the properties in the set/remove element */
	for (one_prop = prop->first_child; one_prop;
	     one_prop = one_prop->next) {
	    dav_response r;

	    if (is_remove) {
		r = dav_delete_prop(db, one_prop, pbuf);
	    }
	    else {
		r = dav_set_prop(db, one_prop, pbuf);
	    }
	    if (r.status != HTTP_OK) {
		/* ### this is way wrong. we need to do a rollback and
		 * ### collect a set of dav_response structures so that
		 * ### we can note the failure and then issue a bunch of
		 * ### dependency failures. */
		return r;
	    }

	    dav_text_append(p, &hdr, dav_empty_elem(p, one_prop));
	}
    }

    dav_text_append(p, &hdr,
		    "</D:prop>" DEBUG_CR
		    "<D:status>HTTP/1.1 200 OK</D:status>" DEBUG_CR
		    "</D:propstat>" DEBUG_CR);

    resp.propstats.propstats = hdr.first;
    return resp;
}

/* handle the PROPPATCH method */
static int dav_method_proppatch(request_rec *r)
{
    int result;
    dav_xml_doc *doc;
    const char *dirpath;
    const char *fname;
    dav_xml_elem *child;
    dav_buffer work_buf = { 0 };
    dav_propdb *db;
    int dependency_failure = 0;
    int has_set = 0;
    dav_response resp = { 0 };
    int dberr;

    if (r->finfo.st_mode == 0) {
	/* Apache will supply a default error for this. */
	return HTTP_NOT_FOUND;
    }

    if ((result = dav_parse_input(r, &doc)) != OK) {
	return result;
    }
    /* note: doc == NULL if no request body */

    if (doc == NULL || !dav_validate_root(doc, "propertyupdate")) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "The request body does not contain "
		      "a \"propertyupdate\" element.");
	return HTTP_BAD_REQUEST;
    }

    if (S_ISDIR(r->finfo.st_mode)) {
	dirpath = r->filename;
	fname = NULL;
    }
    else {
	/* ### is this the best way to get a directory and file? */
	dirpath = ap_make_dirstr_parent(r->pool, r->filename);
	fname = r->filename + strlen(dirpath);
    }

    /* ensure that our state directory is present for this directory */
    dav_ensure_state_dir(r->pool, dirpath);

    db = dav_open_propdb(r->pool, dirpath, fname, &r->finfo, 0,
			 doc->namespaces, &dberr);
    if (db == NULL) {
	int err;
	const char *errstr;

	/* Bad, bad, things. */
	/* ### figure out something better to do... */
	/* ### this is probably a permission thing */

	dav_propdb_error(NULL, dberr, &err, &errstr);

	/* ### fix this error message (default vs special) */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "could not open propdb: %s (%d).", errstr, err);

	return HTTP_INTERNAL_SERVER_ERROR;
    }
    /* ### what to do about closing the propdb on server failure? */

    /* ### check for locks */
    /* ### validate "live" properties */

    /* do a first pass to ensure that all "remove" properties exist */
    for (child = doc->root->first_child; child; child = child->next) {
	int is_remove;
	dav_xml_elem *prop;
	dav_xml_elem *one_prop;

	/* Ignore children that are not set/remove */
	if (child->ns != DAV_NS_DAV_ID
	    || (!(is_remove = strcmp(child->name, "remove") == 0)
		&& strcmp(child->name, "set") != 0)) {
	    child->annotation = DAV_ANNO_IGNORE;
	    continue;
	}

	/* make sure that a "prop" child exists for set/remove */
	if ((prop = dav_find_child(child, "prop")) == NULL) {
	    dav_close_propdb(db);

	    /* This supplies additional information for the default message. */
	    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			  "A \"prop\" element is missing inside "
			  "the propertyupdate command.");
	    return HTTP_BAD_REQUEST;
	}

	/* we're done if this isn't a "remove" command */
	if (!is_remove) {
	    has_set = 1;
	    continue;
	}
	child->annotation = DAV_ANNO_DAVTAG_REMOVE;

	for (one_prop = prop->first_child; one_prop;
	     one_prop = one_prop->next) {
	    if (!dav_prop_exists(db, one_prop, &work_buf)) {
		one_prop->annotation = DAV_ANNO_PROPPATCH_404;
		dependency_failure = 1;
	    }
	}
    }

    /* ### should test that we found at least one set/remove */

    /* generate a failure response, or get the work done */
    if (dependency_failure) {
	resp = dav_failed_proppatch(r->pool, doc);
    }
    else {
	if (has_set) {
	    /* ### temporary hack to prep the propdb namespace mapping */
	    dav_prop_ctx ctx = { 0 };
	    ctx.db = db;
	    ctx.operation = DAV_PROP_OP_SET;
	    dav_prop_validate(&ctx);
	}

	resp = dav_exec_proppatch(r->pool, doc, db, &work_buf);
    }

    /* make sure this gets closed! */
    dav_close_propdb(db);

    /* eek... something very bad may have happened */
    if (resp.status == HTTP_INTERNAL_SERVER_ERROR) {
	/* ### fix this message: notify that the server state is wonky */

	/* Apache will supply a default error for this. */
	return HTTP_INTERNAL_SERVER_ERROR;
    }

    /* ### probably want something better than unparsed_uri */
    /* ### probably need to escape this thing in some way */
    resp.href = r->unparsed_uri;

    /* all PROPPATCH results are "207 Multi-Status" and of type text/xml */
    r->status = HTTP_MULTI_STATUS;
    r->content_type = DAV_XML_CONTENT_TYPE;

    ap_send_http_header(r);

    ap_soft_timeout("send PROPPATCH response", r);

    dav_send_multistatus(r, &resp, doc->namespaces);

#if 0
    if (doc->root) {
	recurse_elem(doc, doc->root, r);
    }
#endif

    ap_kill_timeout(r);

    /* the response has been sent. */
    return DONE;
}

static int process_mkcol_body(request_rec *r)
{
    /* This is snarfed from ap_setup_client_block(). We could get pretty
     * close to this behavior by passing REQUEST_NO_BODY, but we need to
     * return HTTP_UNSUPPORTED_MEDIA_TYPE (while ap_setup_client_block
     * returns HTTP_REQUEST_ENTITY_TOO_LARGE). */

    const char *tenc = ap_table_get(r->headers_in, "Transfer-Encoding");
    const char *lenp = ap_table_get(r->headers_in, "Content-Length");
    int has_body = 0;

    if (tenc) {
	if (strcasecmp(tenc, "chunked")) {
	    /* Use this instead of Apache's default error string */
	    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			  "Unknown Transfer-Encoding %s", tenc);
	    return HTTP_NOT_IMPLEMENTED;
	}

	has_body = 1;
    }
    else if (lenp) {
	const char *pos = lenp;

	while (ap_isdigit(*pos) || ap_isspace(*pos)) {
	    ++pos;
	}
	if (*pos != '\0') {
	    /* This supplies additional information for the default message. */
	    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			  "Invalid Content-Length %s", lenp);
	    return HTTP_BAD_REQUEST;
	}

	has_body = atol(lenp) > 0;
    }

    if (has_body) {
	int result;

	/* ### log something? */

	/* ### I presume we ARE supposed to do this... */
	if ((result = ap_discard_request_body(r)) != OK) {
	    return result;
	}

	/* Apache will supply a default error for this. */
	return HTTP_UNSUPPORTED_MEDIA_TYPE;
    }

    return OK;
}

/* handle the MKCOL method */
static int dav_method_mkcol(request_rec *r)
{
    const char *body;
    int result;
    dav_dir_conf *conf;

    /* handle the request body */
    /* ### this may move lower once we start processing bodies */
    if ((result = process_mkcol_body(r)) != OK) {
	return result;
    }

    conf = (dav_dir_conf *) ap_get_module_config(r->per_dir_config,
						 &dav_module);

    /* ### appropriate lock checks go in here somewhere... */

    if (conf->hooks->mkcol_test) {
	/* ### should this be a standard return value rather than a bool? */
	int proceed = (*conf->hooks->mkcol_test)(r);

	if (!proceed) {
	    /* Apache will supply a default error for this. */
	    return HTTP_METHOD_NOT_ALLOWED;
	}
    }
    else {
	if (DAV_REQ_HAS_EXTRA_PATH_INFO(r)) {
	    /* a parent directory did not exist */

	    /* Supply some text for the error response body. */
	    body = ap_psprintf(r->pool,
			    "Missing one or more intermediate collections. "
			       "Cannot create collection %s.",
			       ap_escape_html(r->pool, r->uri));
	    return dav_error_response(r, HTTP_CONFLICT, NULL, body);
	}

	if (r->finfo.st_mode != 0) {
	    /* oops. something was already there! */

	    /* Apache will supply a default error for this. */
	    return HTTP_METHOD_NOT_ALLOWED;
	}
    }

    if (conf->hooks->mkcol_create) {
	if ((result = (*conf->hooks->mkcol_create)(r)) != OK) {
	    return result;
	}
    }
    else {
	/* ### do we need to deal with the umask? */
	if (mkdir(r->filename, DAV_MODE_DIR) != 0) {
	    /* Apache will supply a default error for this. */
	    /* ### not quite right. we should mention MKCOL */
	    return HTTP_FORBIDDEN;
	}
    }

    /* return an appropriate response (HTTP_CREATED or HTTP_NO_CONTENT) */
    return dav_created(r, NULL, "Collection");
}

static int dav_copymove_walker(dav_walker_ctx *ctx)
{
    dav_response resp;

    /* ### need to completely revise the error handling here to deal
     * ### with multi-status. also need to deal with skipping internal
     * ### member collections. */

    if (ctx->fname == NULL) {
	if (ctx->finfo.st_mode == 0) {
	    /* ### more work here. we may have had to skip a collection */
	    /* postfix call for MOVE. delete the source dir */
	    if (rmdir(ctx->pathbuf.buf) != 0) {
	    }
	}
	/* ### do we need to deal with the umask? */
	else if (mkdir(ctx->path2.buf, DAV_MODE_DIR) != 0) {
	}
    }
    else {
	resp = dav_copymove_file(ctx->is_move, ctx->pool, ctx->pathbuf.buf,
				 ctx->path2.buf, &ctx->work_buf);
	if (resp.status != HTTP_OK) {
	    /* ### this should be linked into a multi-status */
	    if (resp.desc != NULL) {
		errno = resp.save_errno;
		ap_log_rerror(APLOG_MARK, APLOG_ERR, ctx->r, "%s", resp.desc);
	    }
	    return resp.status;
	}
    }

    /* except for postfix calls, also copy/move the propset */
    if (ctx->move_props && (ctx->fname != NULL || ctx->finfo.st_mode != 0)) {
	resp = dav_copymove_propset(ctx->is_move, ctx->pool,
				    ctx->pathbuf.buf, ctx->path2.buf,
				    &ctx->work_buf);
	if (resp.status != HTTP_OK) {
	    /* ### this should be linked into a multi-status */
	    if (resp.desc != NULL) {
		errno = resp.save_errno;
		ap_log_rerror(APLOG_MARK, APLOG_ERR, ctx->r, "%s", resp.desc);
	    }
	    return resp.status;
	}
    }

    return OK;
}

/* ### expand to allow cross-dev moves via dav_copymove_file() */
static int dav_simple_move(request_rec *r, request_rec *rnew)
{
    dav_response resp;
    int result;

    if (rnew->finfo.st_mode != 0) {
	/* target exists; wipe it. We know Overwrite was provided or we
	 * wouldn't have made it this far. */
	if ((result = dav_delete_tree(rnew)) != OK) {
	    return result;
	}
    }

    /* move the file/directory */
    if (rename(r->filename, rnew->filename) != 0) {
	/* ### use a different HTTP status code? */

	/* We will override the default error message */
	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
		      "Could not move resource/collection");
	return HTTP_INTERNAL_SERVER_ERROR;
    }

    /* if a file, then move its associated property set */
    if (S_ISREG(r->finfo.st_mode)) {
	resp = dav_copymove_propset(DAV_DO_MOVE, r->pool, r->filename,
				    rnew->filename, NULL);
	if (resp.status != HTTP_OK) {
	    errno = resp.save_errno;
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s", resp.desc);

	    /* move the original back */
	    if (rename(rnew->filename, r->filename) != 0) {
		/* ### ACK. We should probably do a bit more here. */

		/* We will override the default error message */
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
			      "Could not move file back after a failure to "
			      "move its property set. Repository is now in "
			      "an inconsistent state");
		return HTTP_INTERNAL_SERVER_ERROR;
	    }

	    return resp.status;
	}
    }

    /* return an appropriate response (HTTP_CREATED or HTTP_NO_CONTENT) */
    return dav_created(r, rnew, "Destination");
}

/* copy a single file and its properties */
static int dav_simple_copy(request_rec *r, request_rec *rnew)
{
    dav_response resp;
    int result;
    dav_buffer buf = { 0 };

    /* copy the file itself */
    resp = dav_copymove_file(DAV_DO_COPY, r->pool, r->filename,
			     rnew->filename, &buf);
    if (resp.status != HTTP_OK) {
	if (resp.desc != NULL) {
	    errno = resp.save_errno;
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s", resp.desc);
	}
	result = resp.status;	/* ### change to use only use resp */
	goto error;
    }

    /* now its properties */
    resp = dav_copymove_propset(DAV_DO_COPY, r->pool, r->filename,
				rnew->filename, &buf);
    if (resp.status != HTTP_OK) {
	if (resp.desc != NULL) {
	    errno = resp.save_errno;
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s", resp.desc);
	}
	result = resp.status;	/* ### change to use only use resp */

	/* get rid of the copy of the file */
	if (remove(rnew->filename) != 0) {
	    /* ### crap. what to do now? */

	    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			  "The file was copied, and a failure occurred "
			  "during the copy of its properties. The file "
			  "copy could not be undone. The server is now "
			  "in an inconsistent state.\n\n"
			  "Additional inconsistencies may exist if a "
			  "file was overwritten at the destination -- "
			  "the overwritten file is lost and its properties "
			  "may or may not still exist at the destination.");
	    return HTTP_INTERNAL_SERVER_ERROR;
	}

	/* ### we should ensure the propdb at the destination is deleted. */
	/* ### this could log an error that we don't want */
	resp = dav_delete_propset(r->pool, rnew->filename);
	if (resp.status != HTTP_OK) {
	    if (resp.desc != NULL) {
		errno = resp.save_errno;
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "%s", resp.desc);
	    }
	    /* note: do not override previous stashing of "result" */
	}

	goto error;
    }

    /* return an appropriate response (HTTP_CREATED or HTTP_NO_CONTENT) */
    return dav_created(r, rnew, "Destination");

  error:

    /* see if we overwrote anything */
    if (rnew->finfo.st_mode != 0) {
	/* ### crap. what to do now? */

	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "A file at the destination of the copy was "
		      "overwritten, but then deleted "
		      "after the source file's property set could not be "
		      "copied. "
		      "The overwritten file has been deleted and its "
		      "old properties are in an unknown state. "
		      "The repository is most likely in an inconsistent "
		      "state.");
	return HTTP_INTERNAL_SERVER_ERROR;
    }
    return result;
}

/* handle the COPY and MOVE methods */
static int dav_method_copymove(request_rec *r, int is_move)
{
    const char *body;
    const char *dest;
    dav_lookup_result lookup;
    int is_dir;
    int overwrite;
    int depth;
    int srclen;
    int dstlen;
    dav_walker_ctx ctx = { 0 };
    int result;

    /* source better be there */
    if (r->finfo.st_mode == 0) {
	/* Apache will supply a default error for this. */
	return HTTP_NOT_FOUND;
    }

    dest = ap_table_get(r->headers_in, "Destination");
    if (dest == NULL) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		      "The request is missing a Destination header.");
	return HTTP_BAD_REQUEST;
    }

    lookup = dav_lookup_uri(dest, r);
    if (lookup.rnew == NULL) {
	if (lookup.resp.status == HTTP_BAD_REQUEST) {
	    /* This supplies additional information for the default message. */
	    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			  lookup.resp.desc);
	    return HTTP_BAD_REQUEST;
	}

	/* ### this assumes that dav_lookup_uri() only generates a status
	 * ### that Apache can provide a status line for!! */

	return dav_error_response(r, lookup.resp.status, NULL,
				  lookup.resp.desc);
    }
    if (lookup.rnew->status != HTTP_OK) {
	/* ### how best to report this... */
	return dav_error_response(r, lookup.rnew->status, NULL,
				  "Destination URI had an error.");
    }

    if (DAV_REQ_HAS_EXTRA_PATH_INFO(lookup.rnew)) {
	/* a parent directory did not exist */

	/* Supply some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Missing one or more intermediate collections "
			   "for the destination of %s.",
			   ap_escape_html(r->pool, lookup.rnew->uri));
	return dav_error_response(r, HTTP_CONFLICT, NULL, body);
    }

    /* get and parse the overwrite header value */
    if ((overwrite = dav_get_overwrite(r)) < 0) {
	/* dav_get_overwrite() supplies additional information for the
	 * default message. */
	return HTTP_BAD_REQUEST;
    }

    /* quick failure test: if dest exists and overwrite is false. */
    if (lookup.rnew->finfo.st_mode != 0 && !overwrite) {
	/* Supply some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Destination is not empty and "
			   "Overwrite is not \"T\"");
	return dav_error_response(r, HTTP_PRECONDITION_FAILED, NULL, body);
    }

    /* are the source and destination the same? */
    if (lookup.rnew->finfo.st_mode != 0
	&& r->finfo.st_ino == lookup.rnew->finfo.st_ino) {
	/* Supply some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Source and Destination URIs are the same.");
	return dav_error_response(r, HTTP_FORBIDDEN, NULL, body);
    }

    is_dir = S_ISDIR(r->finfo.st_mode);

    /* get and parse the Depth header value. "0" and "infinity" are legal. */
    if ((depth = dav_get_depth(r, DAV_INFINITY)) < 0) {
	/* dav_get_depth() supplies additional information for the
	 * default message. */
	return HTTP_BAD_REQUEST;
    }
    if (depth == 1) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		   "Depth must be \"0\" or \"infinity\" for COPY or MOVE.");
	return HTTP_BAD_REQUEST;
    }
    if (is_move && is_dir && depth != DAV_INFINITY) {
	/* This supplies additional information for the default message. */
	ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		    "Depth must be \"infinity\" when moving a collection.");
	return HTTP_BAD_REQUEST;
    }

    srclen = strlen(r->filename);
    dstlen = strlen(lookup.rnew->filename);

    if (is_dir
	&& depth == DAV_INFINITY
	&& dstlen > srclen
	&& memcmp(r->filename, lookup.rnew->filename, srclen) == 0
	&& lookup.rnew->filename[srclen] == '/') {
	/* Supply some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Source collection contains the Destination.");
	return dav_error_response(r, HTTP_FORBIDDEN, NULL, body);
    }
    if (is_dir
	&& srclen > dstlen
	&& memcmp(r->filename, lookup.rnew->filename, dstlen) == 0
	&& r->filename[dstlen] == '/') {

	/* The destination must exist (since it contains the source), and
	 * a condition above implies Overwrite==T. Obviously, we cannot
	 * delete the Destination before the MOVE/COPY, as that would
	 * delete the Source.
	 */

	/* Supply some text for the error response body. */
	body = ap_psprintf(r->pool,
			   "Destination collection contains the Source and "
			   "Overwrite has been specified.");
	return dav_error_response(r, HTTP_FORBIDDEN, NULL, body);
    }

    /* ### for now, we don't need anything in the body */
    if ((result = ap_discard_request_body(r)) != OK) {
	return result;
    }

    /* for a MOVE, we can shortcut if the destination is on the same
     * device as the source. */
    if (is_move) {
	const char *dirpath;
	struct stat finfo;

	if (lookup.rnew->finfo.st_mode != 0
	    && lookup.rnew->finfo.st_dev == r->finfo.st_dev) {
	    /* target exists and is on the same device. */
	    return dav_simple_move(r, lookup.rnew);
	}

	/* get information on the parent directory of the destination */
	dirpath = ap_make_dirstr_parent(lookup.rnew->pool,
					lookup.rnew->filename);
	if (stat(dirpath, &finfo) != 0) {
	    /* ### use a different HTTP status code? */

	    /* We will override the default error message */
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
			  "The destination directory disappeared");
	    return HTTP_INTERNAL_SERVER_ERROR;
	}
	if (finfo.st_dev == r->finfo.st_dev) {
	    return dav_simple_move(r, lookup.rnew);
	}
    }
    else if (!is_dir) {
	/* need special care for copying single files. */
	return dav_simple_copy(r, lookup.rnew);
    }

    /* ### need to ensure that we delete the target! */

    /* prep the walking context */
    ctx.walk_type = DAV_WALKTYPE_FULL;
    ctx.func = dav_copymove_walker;
    ctx.pool = r->pool;
    ctx.finfo = r->finfo;
    ctx.is_move = is_move;
    ctx.overwrite = overwrite;
    ctx.postfix = is_move;	/* needed for MOVE */
    ctx.move_props = is_move && !is_dir;	/* for single-file moves */

    /* ### temp hack for dav_copymove_walker */
    ctx.r = r;

    /* copy the source filename into the buffer (with null term) */
    /* Note: null term occurs within pad area of buffer */
    dav_set_bufsize(r->pool, &ctx.pathbuf, srclen);
    memcpy(ctx.pathbuf.buf, r->filename, ctx.pathbuf.cur_len + 1);

    /* copy over the URI (with null term) */
    dav_set_bufsize(r->pool, &ctx.uri, strlen(r->uri));
    memcpy(ctx.uri.buf, r->uri, ctx.uri.cur_len + 1);

    /* copy the destination filename into the buffer (with null term) */
    /* Note: null term occurs within pad area of buffer */
    dav_set_bufsize(r->pool, &ctx.path2, dstlen);
    memcpy(ctx.path2.buf, lookup.rnew->filename, ctx.path2.cur_len + 1);

    /* if we have a directory, then ensure the URI has a trailing "/" */
    if (is_dir && ctx.uri.buf[ctx.uri.cur_len - 1] != '/') {
	/* this will fall into the pad area we allocated */
	ctx.uri.buf[ctx.uri.cur_len++] = '/';
	ctx.uri.buf[ctx.uri.cur_len] = '\0';
    }

    if ((result = dav_walker(&ctx, depth)) != OK) {
	return result;
    }

    /* if there are no multi-status records, then simplify the status */
    if (ctx.response == NULL) {
	/* return an appropriate response (HTTP_CREATED or HTTP_NO_CONTENT) */
	return dav_created(r, lookup.rnew, "Destination");
    }

    /* prep the response: "207 Multi-Status" and of type text/xml */
    r->status = HTTP_MULTI_STATUS;
    r->content_type = DAV_XML_CONTENT_TYPE;

    ap_send_http_header(r);

    ap_soft_timeout("send COPY/MOVE response", r);

    dav_send_multistatus(r, ctx.response, NULL);

#if 0
    if (doc->root) {
	recurse_elem(doc, doc->root, r);
    }
#endif

    ap_kill_timeout(r);

    /* the response has been sent. */
    return DONE;
}

/* handle the LOCK method */
static int dav_method_lock(request_rec *r)
{
    /* ### we know this method, but we won't allow it yet (NYI) */
    /* ### also: no locks yet */
    /* Apache will supply a default error for this. */
    return HTTP_METHOD_NOT_ALLOWED;
}

/* handle the UNLOCK method */
static int dav_method_unlock(request_rec *r)
{
    /* ### we know this method, but we won't allow it yet (NYI) */
    /* ### also: no locks yet */
    /* Apache will supply a default error for this. */
    return HTTP_METHOD_NOT_ALLOWED;
}


/* handle the SEARCH method from DASL */
static int dav_method_search(request_rec *r)
{
    /* ### we know this method, but we won't allow it yet */
    /* Apache will supply a default error for this. */
    return HTTP_METHOD_NOT_ALLOWED;

    /* Do some error checking, like if the querygrammar is
     * supported by the content type, and then pass the
     * request on to the appropriate query module.
     */
}


/*
 * Response handler for DAV resources
 */
static int dav_handler(request_rec *r)
{
    dav_dir_conf *conf;
    int result;

    conf = (dav_dir_conf *) ap_get_module_config(r->per_dir_config,
						 &dav_module);

    /* quickly ignore any HTTP/0.9 requests */
    if (r->assbackwards) {
	return DECLINED;
    }

    /* ### do we need to do anything with r->proxyreq ?? */

    /* NOTE: M_GET would be needed if we enabled read-locks or ACLs */
    /* ### more work necessary, now that we have M_foo for DAV methods */
    /*
     * Set up the methods mask, since that's one of the reasons this handler
     * gets called, and lower-level things may need the info.
     *
     * First, set the mask to the methods we handle directly.  Since by
     * definition we own our managed space, we unconditionally set
     * the r->allowed field rather than ORing our values with anything
     * any other module may have put in there.
     *
     * These are the HTTP-defined methods that we handle directly.
     */
    r->allowed =
	(1 << M_PUT) |
	(1 << M_DELETE) |
	(1 << M_OPTIONS) |
	(1 << M_INVALID);
    /*
     * These are the DAV methods we handle.
     */
    r->allowed |= 0
	| (1 << M_COPY)
	| (1 << M_MKCOL)
	| (1 << M_MOVE)
	| (1 << M_PROPFIND)
	| (1 << M_PROPPATCH);
    /*
     * These are methods that we don't handle directly, but let the
     * server's default handler do for us as our agent.
     */
    r->allowed |= 0
	| (1 << M_GET)
	| (1 << M_POST);
 
    /* ### hrm. if we return HTTP_METHOD_NOT_ALLOWED, then an Allow header
     * ### is sent; it will need the other allowed states; since the default
     * ### handler is not called on error, then it doesn't add the other
     * ### allowed states, so we must */
    /* ### we might need to refine this for just where we return the error.
     * ### also, there is the issue with other methods (see ISSUES) */
    /* ### more work necessary, now that we have M_foo for DAV methods */

    /* this hook allows a subsystem to initialize itself */
    if (conf->hooks->request_handler) {
	if ((result = (*conf->hooks->request_handler)(r)) != OK) {
	    return result;
	}
    }

    /* dispatch the appropriate method handler */
    if (r->method_number == M_PUT) {
	return dav_method_put(r);
    }

    if (r->method_number == M_DELETE) {
	return dav_method_delete(r);
    }

    if (r->method_number == M_OPTIONS) {
	return dav_method_options(r);
    }

    if (r->method_number == M_PROPFIND) {
	return dav_method_propfind(r);
    }

    if (r->method_number == M_PROPPATCH) {
	return dav_method_proppatch(r);
    }

    if (r->method_number == M_MKCOL) {
	return dav_method_mkcol(r);
    }

    if (r->method_number == M_COPY) {
	return dav_method_copymove(r, DAV_DO_COPY);
    }

    if (r->method_number == M_MOVE) {
	return dav_method_copymove(r, DAV_DO_MOVE);
    }

    if (r->method_number == M_LOCK) {
	return dav_method_lock(r);
    }

    if (r->method_number == M_UNLOCK) {
	return dav_method_unlock(r);
    }

    /*
     * NOTE: When Apache moves creates defines for the add'l DAV methods,
     *       then it will no longer use M_INVALID. This code must be
     *       updated each time Apache adds method defines.
     */
    if (r->method_number != M_INVALID) {
	return DECLINED;
    }

    if (!strcmp(r->method, "SEARCH")) {
	return dav_method_search(r);
    }

    /* ### add'l methods for Advanced Collections, ACLs, Versioning, DASL */

    return DECLINED;
}

static int dav_type_checker(request_rec *r)
{
    dav_dir_conf *conf;

    conf = (dav_dir_conf *) ap_get_module_config(r->per_dir_config,
						 &dav_module);
    /* if DAV is enabled, and this isn't a GET or POST, then use our handler */
    /* ### we should (instead) trap the ones that we DO understand */
    if (conf->enabled
	&& r->method_number != M_GET
	&& r->method_number != M_POST) {
	return dav_api_type_checker(r);
    }

    return DECLINED;
}

int dav_api_type_checker(request_rec *r)
{
    /* ### anything else to do here? could another module and/or config
     * ### option "take over" the handler here? i.e. how do we lock down
     * ### this hierarchy so that we are the ultimate arbiter?
     * ### (or do we simply depend on the administrator to avoid
     * ### conflicting configurations?)
     */

    r->handler = "dav-handler";
    return OK;
}

void dav_api_set_request_handler(request_rec *r,
				 dav_api_request_handler handler)
{
    dav_dir_conf *conf;

    conf = (dav_dir_conf *) ap_get_module_config(r->per_dir_config,
						 &dav_module);
    conf->hooks->request_handler = handler;
}

void dav_api_set_mkcol_handlers(request_rec *r,
				dav_api_request_handler test,
				dav_api_request_handler create)
{
    dav_dir_conf *conf;

    conf = (dav_dir_conf *) ap_get_module_config(r->per_dir_config,
						 &dav_module);
    conf->hooks->mkcol_test = test;
    conf->hooks->mkcol_create = create;
}

int dav_api_error_response(request_rec *r, int status, const char *sline,
			   const char *body)
{
    return dav_error_response(r, status, sline, body);
}

/*---------------------------------------------------------------------------
**
** Configuration info for the module
*/

static const command_rec dav_cmds[] =
{
    {
	"DAV",
	dav_cmd_dav,
	NULL,
	ACCESS_CONF,
	FLAG,
	"turn DAV on/off for a directory or location"
    },
    {
	"AddQueryGrammar",
	dav_cmd_addquerygrammar,
	NULL,
	RSRC_CONF,
	TAKE3,
	"AddQueryGrammar <grammar name> <module name> <coded uri>"
    },
    {
	"AllowQueryGrammar",
	dav_cmd_allowquerygrammar,
	NULL,
	RSRC_CONF,
	ITERATE2,
    "AllowQueryGrammar <Content-type> <query grammar> [<query grammar> ...]"
    },
    {NULL}
};

static const handler_rec dav_handlers[] =
{
    {"dav-handler", dav_handler},
    {NULL}
};

module MODULE_VAR_EXPORT dav_module =
{
    STANDARD_MODULE_STUFF,
    dav_init_handler,		/* initializer */
    dav_create_dir_config,	/* dir config creater */
    dav_merge_dir_config,	/* dir merger --- default is to override */
    dav_create_server_config,	/* server config */
    dav_merge_server_config,	/* merge server config */
    dav_cmds,			/* command table */
    dav_handlers,		/* handlers */
    NULL,			/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    dav_type_checker,		/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
