/*
** Copyright (C) 1998-1999 Greg Stein. All Rights Reserved.
**
** By using this file, you agree to the terms and conditions set forth in
** the LICENSE.html file which can be found at the top level of the mod_dav
** distribution or at http://www.webdav.org/mod_dav/license-1.html.
**
** Contact information:
**   Greg Stein, PO Box 3151, Redmond, WA, 98073
**   gstein@lyra.org, http://www.webdav.org/mod_dav/
*/

/*
** DAV extension module for Apache 1.3.*
**
** Writen by Greg Stein, gstein@lyra.org, http://www.lyra.org/
*/

#ifndef _MOD_DAV_H_
#define _MOD_DAV_H_

#include "httpd.h"


#define DAV_VERSION		"0.9.8"

/* the subdirectory to hold all DAV-related information for a directory */
#define DAV_STATE_DIR		".DAV"
#define DAV_STATE_FILE_FOR_DIR	".state_for_dir"

#define DAV_XML_HEADER		"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
#define DAV_XML_CONTENT_TYPE	"text/xml; charset=\"utf-8\""

#define DAV_READ_BLOCKSIZE	2048	/* used for reading input blocks */
#define DAV_COPY_BLOCKSIZE	16384	/* copy 16k at a time */

#define DAV_MODE_DIR		(S_IRWXU | S_IRWXG)
#define DAV_MODE_FILE		(S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)

#define DAV_RESPONSE_BODY_1	"<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n<HTML><HEAD>\n<TITLE>"
#define DAV_RESPONSE_BODY_2	"</TITLE>\n</HEAD><BODY>\n<H1>"
#define DAV_RESPONSE_BODY_3	"</H1>\n"
#define DAV_RESPONSE_BODY_4	"</BODY></HTML>\n"

#define DAV_DO_COPY		0
#define DAV_DO_MOVE		1


#if 1
#define DEBUG_CR	"\n"
#else
#define DEBUG_CR	""
#endif


/* --------------------------------------------------------------------
**
** HANDY UTILITIES
*/

/* simple strutures to keep a linked list of pieces of text */
typedef struct dav_text
{
    const char *text;
    struct dav_text *next;
} dav_text;

typedef struct
{
    dav_text *first;
    dav_text *last;
} dav_text_header;

/* buffer for reuse; can grow to accomodate needed size */
typedef struct
{
    int alloc_len;
    int cur_len;
    char *buf;
} dav_buffer;
#define DAV_BUFFER_MINSIZE	256	/* minimum size for buffer */
#define DAV_BUFFER_PAD		64	/* amount of pad when growing */

/* contains results from one of the getprop functions */
typedef struct
{
    dav_text * propstats;	/* <propstat> element text */
    dav_text * xmlns;		/* namespace decls for <response> elem */
} dav_get_props_result;

/* holds the contents of a <response> element */
typedef struct dav_response
{
    const char *href;		/* always */
    const char *desc;		/* optional description */

    /* use one of these two. use status if propstats.propstats is NULL. */
    int status;
    dav_get_props_result propstats;

    int save_errno;		/* optional copy of errno causing error */

    struct dav_response *next;
} dav_response;

typedef struct
{
    request_rec *rnew;		/* new subrequest */
    dav_response resp;		/* potential error response */
} dav_lookup_result;


void dav_text_append(pool *p, dav_text_header *hdr, const char *text);

void dav_check_bufsize(pool *p, dav_buffer *pbuf, int extra_needed);
void dav_set_bufsize(pool *p, dav_buffer *pbuf, int size);

dav_lookup_result dav_lookup_uri(const char *uri, request_rec *r);

void dav_ensure_state_dir(pool *p, const char *dirname);

/* copy a file (specified with full paths). possibly delete after the copy */
dav_response dav_copymove_file(
    int is_move,
    pool *p,
    const char *src, const char *dst,
    dav_buffer *pbuf);

/* move a file from within a state directory to another state directory */
dav_response dav_copymove_state(
    int is_move,
    pool *p,
    const char *src_dir, const char *src_file,
    const char *dst_dir, const char *dst_file,
    dav_buffer *pbuf);


/* --------------------------------------------------------------------
**
** XML PARSING
*/

/*
** Qualified namespace values
**
** DAV_NS_DAV_ID
**    We always insert the "DAV:" namespace URI at the head of the
**    namespace array. This means that it will always be at ID==0,
**    making it much easier to test for.
**
** DAV_NS_NONE
**    This special ID is used for two situations:
**
**    1) The namespace prefix begins with "xml" (and we do not know
**       what it means). Namespace prefixes with "xml" (any case) as
**       their first three characters are reserved by the XML Namespaces
**       specification for future use. mod_dav will pass these through
**       unchanged. When this identifier is used, the prefix is LEFT in
**       the element/attribute name. Downstream processing should not
**       prepend another prefix.
**
**    2) The element/attribute does not have a namespace.
**
**       a) No prefix was used, and a default namespace has not been
**          defined.
**       b) No prefix was used, and the default namespace was specified
**          to mean "no namespace". This is done with a namespace
**          declaration of:  xmlns=""
**          (this declaration is typically used to override a previous
**          specification for the default namespace)
**
**       In these cases, we need to record that the elem/attr has no
**       namespace so that we will not attempt to prepend a prefix.
**       All namespaces that are used will have a prefix assigned to
**       them -- mod_dav will never set or use the default namespace
**       when generating XML. This means that "no prefix" will always
**       mean "no namespace".
**
**    In both cases, the XML generation will avoid prepending a prefix.
**    For the first case, this means the original prefix/name will be
**    inserted into the output stream. For the latter case, it means
**    the name will have no prefix, and since we never define a default
**    namespace, this means it will have no namespace.
**
** Note: currently, mod_dav understands the "xmlns" prefix and the
**     "xml:lang" attribute. These are handled specially (they aren't
**     left within the XML tree), so the DAV_NS_NONE value won't ever
**     really apply to these values.
*/
#define DAV_NS_DAV_ID		0	/* namespace ID for "DAV:" */
#define DAV_NS_NONE		-10	/* no namespace for this elem/attr */

#define DAV_NS_ERROR_BASE	-100	/* used only during processing */
#define DAV_NS_IS_ERROR(e)	((e) <= DAV_NS_ERROR_BASE)


/*
** dav_xml_doc: holds a parsed XML document
** dav_xml_elem: holds a parsed XML element
** dav_xml_attr: holds a parsed XML attribute
**
** dav_xml_ns_scope: internal struct used during processing to scope
**                   namespace declarations
*/

typedef struct dav_xml_attr
{
    const char *name;			/* attribute name */
    int ns;				/* index into namespace array */

    const char *value;			/* attribute value */

    struct dav_xml_attr *next;		/* next attribute */
} dav_xml_attr;

typedef struct dav_xml_elem
{
    const char *name;			/* element name */
    int ns;				/* index into namespace array */
    const char *lang;			/* xml:lang for attrs/contents */

    dav_text_header first_cdata;	/* cdata right after start tag */
    dav_text_header following_cdata;	/* cdata after MY end tag */

    struct dav_xml_elem *parent;	/* parent element */
    struct dav_xml_elem *next;		/* next (sibling) element */
    struct dav_xml_elem *first_child;	/* first child element */
    struct dav_xml_attr *attr;		/* first attribute */

    /* used only during parsing */
    struct dav_xml_elem *last_child;	/* last child element */
    struct dav_xml_ns_scope *ns_scope;	/* namespaces scoped by this elem */

    /* used during request processing */
    int annotation;			/* info about this element */
#define DAV_ANNO_PROPPATCH_404	1	/* this prop was not found */
#define DAV_ANNO_IGNORE		2	/* ignore this element */
#define DAV_ANNO_DAVTAG_REMOVE	3	/* elem tag is "remove" */

#define DAV_ANNO_DAVPROP_START	100	/* first dav property annotation */
#define DAV_ANNO_DAVPROP_END	199	/* last dav property annotation */

} dav_xml_elem;

#define DAV_ELEM_IS_EMPTY(e)	((e)->first_child == NULL && \
				 (e)->first_cdata.first == NULL)

typedef struct dav_xml_doc
{
    dav_xml_elem *root;			/* root element */

    array_header *namespaces;		/* array of namespaces used */
#define DAV_GET_NS_URI(ns_array, i)	(((const char **)(ns_array)->elts)[i])

} dav_xml_doc;


int dav_parse_input(request_rec *r, dav_xml_doc **pdoc);

int dav_validate_root(const dav_xml_doc *doc, const char *tagname);

dav_xml_elem *dav_find_child(
    const dav_xml_elem *elem,
    const char *tagname);

void dav_xml2text(
    pool *p,
    const dav_xml_elem *elem,
    int style,
    int *ns_map,
    const char **pbuf,
    int *psize
    );
#define DAV_X2T_FULL		0	/* start tag, contents, end tag */
#define DAV_X2T_INNER		1	/* contents only */
#define DAV_X2T_LANG_INNER	2	/* xml:lang + inner contents */

const char *dav_empty_elem(pool *p, const dav_xml_elem *elem);


/* --------------------------------------------------------------------
**
** PROPERTY HANDLING
*/

typedef struct dav_propdb dav_propdb;


dav_propdb *dav_open_propdb(
    pool *p,
    const char *dirname, const char *fname,
    struct stat *finfo,
    int ro,
    array_header *ns_xlate,
    int *dberr);

void dav_close_propdb(dav_propdb *db);

void dav_propdb_error(
    dav_propdb *db,
    int dberr,
    int *err, const char **errstr);

dav_get_props_result dav_get_props(
    dav_propdb *db,
    dav_xml_doc *doc);

dav_get_props_result dav_get_allprops(
    dav_propdb *db,
    int getvals);

int dav_prop_exists(
    dav_propdb *db,
    dav_xml_elem *prop,
    dav_buffer *pbuf);

dav_response dav_set_prop(
    dav_propdb *db,
    dav_xml_elem *prop,
    dav_buffer *pbuf);

dav_response dav_delete_prop(
    dav_propdb *db,
    dav_xml_elem *prop,
    dav_buffer *pbuf);

dav_response dav_delete_propset(pool *p, const char *pathname);

dav_response dav_copymove_propset(
    int is_move,
    pool *p,
    const char *src, const char *dst,
    dav_buffer *pbuf);

/*
** 3-phase property modification.
**
**   1) validate props. readable? unlocked? ACLs allow access?
**   2) execute operation (set/delete)
**   3) commit or rollback
**
** ### eventually, auth must be available. a ref to the request_rec (which
** ### contains the auth info) should be in the shared context struct.
**
** Each function may alter the error values and information contained within
** the response record. This should be done as an "increasing" level of
** error, rather than overwriting any previous error.
**
** Note that commit() cannot generate errors. It should simply free the
** rollback information.
**
** rollback() may generate additional errors because the rollback operation
** can sometimes fail(!).
**
** The caller should allocate an array of these, one per operation. Note
** that the set/delete operations are order-dependent.
*/
typedef struct dav_prop_ctx {
    dav_propdb *db;

    int operation;
#define DAV_PROP_OP_SET		1	/* set a property value */
#define DAV_PROP_OP_DELETE	2	/* delete a prop value */
/* ### add a GET? */

    dav_xml_elem *prop;		/* property to affect */

    dav_response response;	/* response for this operation */

    struct dav_rollback_item *rollback;
				/* optional (private) rollback info */
} dav_prop_ctx;

void dav_prop_validate(dav_prop_ctx *ctx);
void dav_prop_exec(dav_prop_ctx *ctx, dav_buffer *pbuf);
void dav_prop_commit(dav_prop_ctx *ctx);
void dav_prop_rollback(dav_prop_ctx *ctx);


/* --------------------------------------------------------------------
**
** HOOK/HANDLER FUNCTIONS
*/

typedef int (*dav_api_request_handler)(request_rec *r);
typedef int (*dav_api_method_handler)(
    char *method,
    table *headers_in,
    dav_xml_doc *body);

/* general APIs for third-parties */
int dav_api_type_checker(request_rec *r);
int dav_api_error_response(request_rec *, int, const char *, const char *);

/* functions for third-parties to set the hooks */
void dav_api_set_request_handler(request_rec *r,
				 dav_api_request_handler handler);
void dav_api_set_mkcol_handlers(request_rec *r,
				dav_api_request_handler test,
				dav_api_request_handler create);


/* --------------------------------------------------------------------
**
** DATABASE FUNCTIONS
*/

typedef struct dav_db dav_db;
typedef struct
{
    char *dptr;
    int dsize;
} dav_datum;

/* hook functions to enable pluggable databases */
typedef struct
{
    int (*open)(
	pool *p,
	const char *dirpath, const char *fname,
	int ro,
	dav_db **pdb);
    void (*close)(dav_db *db);

    int (*fetch)(dav_db *db, dav_datum key, dav_datum *pvalue);
    int (*store)(dav_db *db, dav_datum key, dav_datum value);
    int (*delete)(dav_db *db, dav_datum key);

    int (*exists)(dav_db *db, dav_datum key);

    int (*firstkey)(dav_db *db, dav_datum *pkey);
    int (*nextkey)(dav_db *db, dav_datum *pkey);

    int (*geterror)(dav_db *db);
    const char * (*strerror)(dav_db *db);

    void (*freedatum)(dav_db *db, dav_datum data);

    dav_response (*deleteset)(pool *p, const char *dirpath, const char *fname);
    dav_response (*copymoveset)(
	int is_move,
	pool *p,
	const char *src_dir, const char *src_file,
	const char *dst_dir, const char *dst_file,
	dav_buffer *pbuf);

} dav_hooks_db;

/*
** Error codes returned by the database functions
**
** If the underlying database returns an error, it is expected that the
** dav_db will hold onto it, and return an appropriate error string
** through strerror.
*/
#define DAV_DBERR_NONE		0	/* no error */
#define DAV_DBERR_INTERNAL	1	/* error from internal database */
#define DAV_DBERR_BAD_MAJOR	2	/* major version was wrong */
#define DAV_DBERR_NO_EXIST	3	/* returned by exist() */

/* builtin DBM-style database hooks */
extern const dav_hooks_db dav_hooks_db_dbm;


#endif /* _MOD_DAV_H_ */
